"""
_ML.py
====================================
Main class extension file for the Machine Learning functionality
"""
from copy import deepcopy
import os
import pickle as pkl

from tqdm import tqdm
from tqdm.notebook import tqdm as tqdm_notebook

from .Input_Tools import is_value_subset, is_values_subset
from ..helpers.Data_Helpers import (get_unique_combo_df,
                                    reverse_unique_combo_df,
                                    get_target_type)
from ..helpers.ML_Helpers import (compute_micro_macro, conv_to_list,
                                  get_avaliable_run_name)
from ..pipeline.Evaluator import Evaluator
from ..main.Params_Classes import (CV_Splits, CV_Split, Feat_Importance,
                                   Model_Pipeline,
                                   Model, Ensemble, Problem_Spec)
from ..pipeline.Model_Pipeline import get_pipe
import pandas as pd
import copy


def Set_Default_ML_Verbosity(
 self, save_results='default', progress_bar='default',
 progress_loc='default',
 pipeline_verbose='default',
 best_params_score='default',
 compute_train_score='default',
 show_init_params='default', fold_name='default',
 time_per_fold='default', score_per_fold='default', fold_sizes='default',
 best_params='default', save_to_logs='default', flush='default'):
    '''This function allows setting various verbosity options that effect
    output during :func:`Evaluate` and :func:`Test`.

    Parameters
    ----------
    save_results : bool, optional
        If True, all results returned by Evaluate
        will be saved within the log dr (if one exists!),
        under run_name + .eval, and simmilarly for results
        returned by Test, but as run_name + .test.

        if 'default', and not already defined, set to False.

        ::

            default = 'default'

    progress_bar : bool, optional
        If True, a progress bar, implemented in the python
        library tqdm, is used to show progress during use of
        :func:`Evaluate` , If False, then no progress bar is shown.
        This bar should work both in a notebook env and outside one,
        assuming self.notebook has been set correctly.

        if 'default', and not already defined, set to True.

        ::

            default = 'default'

    progress_loc : str, Path or None, optional
        If not None, then this will record the progress
        of each Evaluate / Test call in this location.

        if 'default', and not already defined, set to False.

        ::

            default = 'default'

    pipeline_verbose : bool, optional
        This controls the verbose parameter for the pipeline object itself.
        If set to True, then time elapsed while fitting each step will be
        printed.

        if 'default', and not already defined, set to False.

        ::

            default = 'default'

    compute_train_score : bool, optional
        If True, then metrics/scorers and raw preds will also be
        computed on the training set in addition to just the
        eval or testing set.

        if 'default', and not already defined, set to False.

        ::

            default = 'default'

    show_init_params : bool, optional
        If True, then print/show the parameters used before running
        Evaluate / Test. If False, then don't print the params used.

        if 'default', and not already defined, set to True.

        ::

            default = 'default'

    fold_name : bool, optional
        If True, prints a rough measure of progress via
        printing out the current fold (somewhat redundant with the
        progress bar if used, except if used with other params, e.g.
        time per fold, then it is helpful to have the time printed
        with each fold). If False, nothing is shown.

        if 'default', and not already defined, set to False.

        ::

            default = 'default'

    time_per_fold : bool, optional
        If True, prints the full time that a fold took to complete.

        if 'default', and not already defined, set to False.

        ::

            default = 'default'

    score_per_fold : bool, optional
        If True, displays the score for each fold, though slightly less
        formatted then in the final display.

        if 'default', and not already defined, set to False.

        ::

            default = 'default'

    fold_sizes : bool, optional
        If True, will show the number of subjects within each train
        and val/test fold.

        if 'default', and not already defined, set to False.

        ::

            default = 'default'

    best_params : bool, optional
        If True, print the best search params found after every
        param search.

        if 'default', and not already defined, set to False.

        ::

            default = 'default'

    save_to_logs : bool, optional
        If True, then when possible, and with the selected model
        verbosity options, verbosity ouput will be saved to the
        log file.

        if 'default', and not already defined, set to False.

        ::

            default = 'default'

    flush : bool, optional
        If True, then add flush=True to all ML prints, which
        adds a call to flush the std output.

        if 'default', and not already defined, set to False.

        ::

            default = False
    '''

    if save_results != 'default':
        self.default_ML_verbosity['save_results'] = save_results
    elif 'save_results' not in self.default_ML_verbosity:
        self.default_ML_verbosity['save_results'] = False

    if progress_bar != 'default':
        if progress_bar is True:
            if self.notebook:
                self.default_ML_verbosity['progress_bar'] = tqdm_notebook
            else:
                self.default_ML_verbosity['progress_bar'] = tqdm
        else:
            self.default_ML_verbosity['progress_bar'] = None
    elif 'progress_bar' not in self.default_ML_verbosity:
        if self.notebook:
            self.default_ML_verbosity['progress_bar'] = tqdm_notebook
        else:
            self.default_ML_verbosity['progress_bar'] = tqdm

    if progress_loc != 'default':
        self.default_ML_verbosity['progress_loc'] = progress_loc
    elif 'progress_loc' not in self.default_ML_verbosity:
        self.default_ML_verbosity['progress_loc'] = None

    if pipeline_verbose != 'default':
        self.default_ML_verbosity['pipeline_verbose'] = pipeline_verbose
    elif 'pipeline_verbose' not in self.default_ML_verbosity:
        self.default_ML_verbosity['pipeline_verbose'] = False

    if compute_train_score != 'default':
        self.default_ML_verbosity['compute_train_score'] = compute_train_score
    elif 'compute_train_score' not in self.default_ML_verbosity:
        self.default_ML_verbosity['compute_train_score'] = False

    if show_init_params != 'default':
        self.default_ML_verbosity['show_init_params'] = show_init_params
    elif 'show_init_params' not in self.default_ML_verbosity:
        self.default_ML_verbosity['show_init_params'] = True

    if fold_name != 'default':
        self.default_ML_verbosity['fold_name'] = fold_name
    elif 'fold_name' not in self.default_ML_verbosity:
        self.default_ML_verbosity['fold_name'] = False

    if time_per_fold != 'default':
        self.default_ML_verbosity['time_per_fold'] = time_per_fold
    elif 'time_per_fold' not in self.default_ML_verbosity:
        self.default_ML_verbosity['time_per_fold'] = False

    if score_per_fold != 'default':
        self.default_ML_verbosity['score_per_fold'] = score_per_fold
    elif 'score_per_fold' not in self.default_ML_verbosity:
        self.default_ML_verbosity['score_per_fold'] = False

    if fold_sizes != 'default':
        self.default_ML_verbosity['fold_sizes'] = fold_sizes
    elif 'fold_sizes' not in self.default_ML_verbosity:
        self.default_ML_verbosity['fold_sizes'] = False

    if best_params != 'default':
        self.default_ML_verbosity['best_params'] = best_params
    elif 'best_params' not in self.default_ML_verbosity:
        self.default_ML_verbosity['best_params'] = False

    if best_params_score != 'default':
        self.default_ML_verbosity['best_params_score'] = best_params
    elif 'best_params_score' not in self.default_ML_verbosity:
        self.default_ML_verbosity['best_params_score'] = False

    if save_to_logs != 'default':
        self.default_ML_verbosity['save_to_logs'] = save_to_logs
    elif 'save_to_logs' not in self.default_ML_verbosity:
        self.default_ML_verbosity['save_to_logs'] = False

    if flush != 'default':
        self.default_ML_verbosity['flush'] = flush
    elif 'flush' not in self.default_ML_verbosity:
        self.default_ML_verbosity['flush'] = False

    self._print('Default ML verbosity set within self.default_ML_verbosity.')
    self._print('----------------------')
    for param in self.default_ML_verbosity:

        if param == 'progress_bar':
            if self.default_ML_verbosity[param] is None:
                self._print(param + ':', False)
            else:
                self._print(param + ':', True)

        else:
            self._print(param + ':', self.default_ML_verbosity[param])

    self._print()


def _ML_print(self, *args, **kwargs):
    '''Overriding the print function to allow for
    customizable verbosity. This print is setup with specific
    settings for the Model_Pipeline class, for using Evaluate and Test.

    Parameters
    ----------
    args
        Anything that would be passed to default python print
    '''

    if self.default_ML_verbosity['save_to_logs']:
        _print = self._print
    else:
        _print = print

    # If flush specified, add as kwarg
    if self.default_ML_verbosity['flush']:
        kwargs['flush'] = True

    level = kwargs.pop('level', None)

    # If no level passed, always print
    if level is None:
        _print(*args, **kwargs)

    elif level == 'name' and self.default_ML_verbosity['fold_name']:
        _print(*args, **kwargs)

    elif level == 'time' and self.default_ML_verbosity['time_per_fold']:
        _print(*args, **kwargs)

    elif level == 'score' and self.default_ML_verbosity['score_per_fold']:
        _print(*args, **kwargs)

    elif level == 'size' and self.default_ML_verbosity['fold_sizes']:
        _print(*args, **kwargs)

    elif level == 'params' and self.default_ML_verbosity['best_params']:
        _print(*args, **kwargs)

    elif level == 'cv_score' and self.default_ML_verbosity[
                                 'best_params_score']:
        _print(*args, **kwargs)


def Evaluate(self,
             model_pipeline,
             problem_spec='default',
             splits=3,
             n_repeats=2,
             cv='default',
             train_subjects='train',
             feat_importances=None,
             return_raw_preds=False,
             return_models=False,
             run_name='default',
             only_fold=None,
             base_dtype='float32',
             CV='depreciated'):
    ''' The Evaluate function is one of the main interfaces
    for building and evaluating :class:`Model_Pipeline` on the loaded data.
    Specifically, Evaluate is designed to try and estimate the out of sample
    performance of a passed :class:`Model_Pipeline` on a specific
    ML task (as specified by :class:`Problem_Spec`).
    This estimate is done through a defined CV strategy
    (`splits` and `n_repeats`). While Evaluate's ideal usage is
    an expirimental context for exploring
    different choices of :class:`Model_Pipeline` and then ultimately
    with :func:`Test<BPt_ML.Test>` -
    if used carefully (i.e., dont try 50 Pipelines's and only report
    the one that does best), it can be used
    on a full dataset.

    Parameters
    ------------
    model_pipeline : :class:`Model_Pipeline`

        The passed `model_pipeline` should be an instance of the BPt params
        class :class:`Model_Pipeline`.
        This object defines the underlying model pipeline to be evaluated.

        See :class:`Model_Pipeline` for more information /
        how to create a the model pipeline.

    problem_spec : :class:`Problem_Spec` or 'default', optional

        `problem_spec` accepts an instance of the BPt.BPt_ML
        params class :class:`Problem_Spec`.
        This object is essentially a wrapper around commonly used
        parameters needs to define the context
        the model pipeline should be evaluated in.
        It includes parameters like problem_type, scorer, n_jobs,
        random_state, etc...
        See :class:`Problem_Spec` explicitly for more information
        and for how to create an instance of this object.

        If left as 'default', then will just initialize a
        Problem_Spec with default params.

        ::

            default = 'default'

    splits : int, float, str or list of str, optional
        In every fold of the defined CV strategy, the passed `model_pipeline`
        will be fitted on
        a train fold, and evaluated on a validation fold. This parameter
        controls the type of CV, i.e., specifies what the train and validation
        folds should be. These splits are further determined by the subjects
        passed to `train_subjects`.
        Notably, the splits defined will respect any special split behavior
        as defined in
        :func:`Define_Validation_Strategy<BPt_ML.Define_Validation_Strategy>`.

        Specifically, options for split are:

        - int
            The number of k-fold splits to conduct. (E.g., 3 for a
            3-fold CV).

        - float
            Must be 0 < `splits` < 1, and defines a single
            train-test like split,
            with `splits` as the % of the current training data size
            used as a validation/test set.

        - str
            If a str is passed, then it must correspond to a
            loaded Strat variable. In
            this case, a leave-out-group CV will be used according
            to the value of the
            indicated Strat variable (E.g., a leave-out-site CV scheme).

        - list of str
            If multiple str passed, first determine the
            overlapping unique values from
            their corresponing loaded Strat variables,
            and then use this overlapped
            value to define the leave-out-group CV as described above.

        Note that this defines only the base CV strategy,
        and that the following param `n_repeats`
        is optionally used to replicate this base strategy, e.g.,
        for a twice repeated train-test split evaluation.
        Note further that `n_repeats` will work with any of these options,
        but say in the case of
        a leave out group CV, it would be awfully redundant, versus,
        with a passed float value, very reasonable.

        ::

            default = 3

    n_repeats : int, optional
        Given the base CV defined / described in the `splits` param, this
        parameter further controls if the defined train/val splits should
        be repeated
        (w/ different random splits in all cases but the
        leave-out-group passed str option).

        For example, if `n_repeats` is set to 2, and `splits` is 3,
        then a twice repeated 3-fold CV
        will be performed, and results returned with respect to this strategy.

        It can be a good idea to set multiple `n_repeats`
        (assuming enough computation power), as it can
        help you spot cases where you may not have enough training subjects
        to get stable behavior, e.g.,
        say you run a three times repeated 3 fold CV,
        if the mean validation scores from each 3-fold are
        all very close to each other, then you know that
        1 repeat is likely enough. If instead the macro std in
        score (the std from in this case those 3 scores) is high,
        then it indicates you may not have enough subjects
        to get stable results from just one 3-fold CV, and
        that you might want to consider changing some settings.

        ::

            default = 2

    cv : 'default' or CV params object, optional
        If left as default 'default', use the class defined CV behavior
        for the splits, otherwise can pass custom behavior

        ::

            default = 'default'

    train_subjects : :ref:`Subjects`, optional
        This parameter determines the set of training subjects which are
        used in this call to `Evaluate`. Note, this parameter is distinct to
        the `subjects` parameter within :class:`Problem_Spec`, which is
        applied after
        selecting the subset of `train_subjects` specified here.
        These subjects are
        used as the input to `Evaluate`, i.e.,
        so typically any subjects data you want
        to remain untouched (say your global test subjects)
        are considered within `Evaluate`,
        and only those explicitly passed here are.

        By default, this value will be set to the special str indiciator
        'train', which specifies that the full set of globally defined
        training subjects
        (See: :func:`Define_Train_Test_Split`),
        should be used. Other special str indicators
        include 'all' to select all subjects, and 'test'
        to select the test set subjects.

        If `subjects` is passed a str, and that str is not one of the str
        indicators listed above, then it will be interpretted as the location
        of file in which to read subjects from
        (assuming one subjects per line).

        `subjects` may also be a custom array-like of subjects to use.

        See :ref:`Subjects` for how to correctly format input and for
        other special options.

        ::

            default = 'train'

    feat_importances : :class:`Feat_Importance` list of, str or None, optional

        If passed None, by default, no feature importances will be saved.

        Alternatively, one may pass the keyword 'base', to indicate that
        the base feature importances - those automatically calculated
        by base objects (e.g., beta weights from linear models) be
        saved. In this case, the object Feat_Importance('base') will be made.

        Otherwise, for more detailed control provide here either
        a single, or list of
        :class:`Feat_Importance` param objects
        in which to specify what importance values, and with what
        settings should be computed.
        See the base :class:`Feat_Importance` object for more information
        on how to specify
        these objects.

        See :ref:`Feat Importances` to learn more about feature importances
        generally.

        In this case of a passed list, all passed Feat_Importances
        will attempt to be
        computed.

        ::

            default = None

    return_raw_preds : bool, optional
        If True, return the raw predictions from each fold.

        ::

            default = False

    return_models : bool, optional
        If True, return the trained models from each evaluation.

        ::

            default = False

    run_name : str or 'default', optional
        Each run of Evaluate can be optionally associated with
        a specific `run_name`. This name
        is used if `save_results` in
        :func:`Set_Default_ML_Verbosity<BPt_ML.Set_Default_ML_Verbosity>`
        is set to True,
        then will be used as the name output from Evaluate as saved as in
        the specific log_dr
        (if any, and as set when Init'ing the
        :class:`BPt_ML <BPt.BPt_ML>` class object),
        with '.eval' appended to the name.

        If left as 'default', will come up with a kind of
        terrible name passed on the underlying
        model used in the passed `model_pipeline`.

        ::

            default = 'default'

    only_fold : int or None, optional
        This is a special parameter used to only
        Evaluate a specific fold of the specified runs to
        evaluate. Keep as None to ignore.

        ::

            default = None

    base_dtype : numpy dtype
        The dataset is cast to a numpy array of float.
        This parameter can be used to change the default
        behavior, e.g., if more resolution or less is needed.

        ::

            default = 'float32'

    CV : 'depreciated'
        Switching to passing cv parameter as cv instead of CV.
        For now if CV is passed it will still work as if it were
        passed as cv.

        ::

            default = 'depreciated'

    Returns
    ----------
    results : dict
        Dictionary containing:
        'summary_scores', A list representation of the
        printed summary scores, where the 0 index is the mean,
        1 index is the macro std, then second index is the micro std.
        'train_summary_scores', Same as summary scores, but only exists
        if train scores are computed.
        'raw_scores', a numpy array of numpy arrays,
        where each internal array contains the raw scores as computed for
        all passed in scorers, computed for each fold within
        each repeat. e.g., array will have a length of `n_repeats` * number of
        folds, and each internal array will have the same length as
        the number of
        scorers. Optionally, this could instead return a list containing as
        the first element the raw training score in this same format,
        and then the raw testing scores.
        'raw_preds', A pandas dataframe containing the raw predictions
        for each subject, in the test set, and
        'FIs' a list where each element corresponds
        to a passed feature importance.

    Notes
    ----------
    Prints by default the following for each scorer,

    float
        The mean macro score (as set by input scorer) across each
        repeated K-fold.

    float
        The standard deviation of the macro score (as set by input scorer)
        across each repeated K-fold.

    float
        The standard deviation of the micro score (as set by input scorer)
        across each fold with the repeated K-fold.

    '''

    # Perform pre-modeling check
    self._premodel_check()

    # Run initial model pipeline check
    model_pipeline = model_pipeline_check(model_pipeline, self.all_data)

    # Should save the params used here*** before any preproc done
    run_name = get_avaliable_run_name(run_name, model_pipeline)

    # Get the the train subjects to use
    _train_subjects = self._get_subjects_to_use(train_subjects)

    # Proc feat importances
    if feat_importances == 'base':
        feat_importances = Feat_Importance(obj='base')
    elif feat_importances == 'default':
        raise RuntimeError('feat_importaces == default is depreciated')

    if CV != 'depreciated':
        print('Warning: Passing CV is depreciated. Please change to',
              'passing as cv instead.')

        # For now, let it still work
        cv = CV

    # Proc. CV
    if cv == 'default':
        cv_obj = self.cv
    else:
        cv_obj = self._get_cv(cv)

    # Pre-proc problem spec, set as copy ps, right before print
    ps = self._preproc_problem_spec(problem_spec)

    # Run checks before print
    model_pipeline._proc_checks()

    # Print the params being used
    if self.default_ML_verbosity['show_init_params']:

        model_pipeline.print_all(self._print)
        ps.print_all(self._print)

        self._print('Evaluate Params')
        self._print('---------------')
        self._print('splits =', splits)
        self._print('n_repeats =', n_repeats)
        self._print('cv =', cv)
        self._print('train_subjects =', train_subjects)
        self._print('feat_importances =', feat_importances)
        self._print('len(train_subjects) =', len(_train_subjects),
                    '(before overlap w/ problem_spec.subjects)')
        self._print('run_name =', run_name)
        self._print()

    # Init the Model_Pipeline object with modeling params
    self._init_evaluator(
        model_pipeline=model_pipeline,
        ps=ps,
        cv=cv_obj,
        feat_importances=feat_importances,
        return_raw_preds=return_raw_preds,
        return_models=return_models,
        base_dtype=base_dtype)

    # Get the Eval splits
    _, splits_vals, _ = self._get_split_vals(splits)

    # Evaluate the model
    train_scores, scores, results =\
        self.evaluator.Evaluate(self.all_data, _train_subjects,
                                splits, n_repeats, splits_vals,
                                only_fold=only_fold)

    # If only fold is not None, set n_repeats = 1
    if only_fold is not None:
        n_repeats = 1

    if 'FIs' in results:
        for fi in results['FIs']:
            fi.set_target(ps.target)
            fi.set_run_name(run_name)

    self._print()

    # Print out summary stats for all passed scorers
    if self.default_ML_verbosity['compute_train_score']:
        score_list = [train_scores, scores]
        score_type_list = ['Training', 'Validation']
    else:
        score_list = [scores]
        score_type_list = ['Validation']

    summary_dfs = {}

    for scrs, name in zip(score_list, score_type_list):

        summary_scores = self._handle_scores(scrs, name,
                                             ps.weight_scorer,
                                             n_repeats, run_name,
                                             self.evaluator.n_splits_,
                                             summary_dfs)

        if name == 'Validation':
            results['summary_scores'] = summary_scores
        else:
            results['train_summary_scores'] = summary_scores

    # Add scores to results
    results['raw_scores'] = score_list

    # Create single scores
    for scorer_str, sum_scores in zip(self.evaluator.scorer_strs,
                                      results['summary_scores']):
        results[scorer_str] = sum_scores[0]

    if 'train_summary_scores' in results:
        for scorer_str, sum_scores in zip(self.evaluator.scorer_strs,
                                          results['train_summary_scores']):
            results['train_' + scorer_str] = sum_scores[0]

    # Saves based on verbose setting
    self._save_results(results, run_name + '.eval')

    return results


def Test(self,
         model_pipeline,
         problem_spec='default',
         train_subjects='train',
         test_subjects='test',
         feat_importances=None,
         return_raw_preds=False,
         return_models=False,
         run_name='default',
         base_dtype='float32'):
    ''' The test function is one of the main interfaces for testing a specific
    :class:`Model_Pipeline`. Test is conceptually different from
    :func:`Evaluate<BPt_ML.Evaluate>`
    in that it is designed to contrust / train a :class:`Model_Pipeline`
    on one discrete set of `train_subjects`
    and evaluate it on a further discrete set of `test_subjects`.
    Otherwise, these functions are very simmilar as
    they both evaluate a :class:`Model_Pipeline` as defined in the context of
    a :class:`Problem_Spec`, and return
    similar output.

    Parameters
    ------------
    model_pipeline : :class:`Model_Pipeline`
        The passed `model_pipeline` should be an instance of the
        BPt params class :class:`Model_Pipeline`.
        This object defines the underlying model pipeline to be evaluated.

        See :class:`Model_Pipeline` for more information / how to
        create a the model pipeline.

    problem_spec : :class:`Problem_Spec` or 'default', optional

        `problem_spec` accepts an instance of the BPt.BPt_ML
        params class :class:`Problem_Spec`.
        This object is essentially a wrapper around commonly used
        parameters needs to define the context
        the model pipeline should be evaluated in.
        It includes parameters like problem_type, scorer, n_jobs,
        random_state, etc...
        See :class:`Problem_Spec` explicitly for more information
        and for how to create an instance of this object.

        If left as 'default', then will just initialize a
        Problem_Spec with default params.

        ::

            default = 'default'

    train_subjects : str, array-like or Value_Subset, optional
        This parameter determines the set of training subjects which are
        used to train the passed instance of :class:`Model_Pipeline`.

        Note, this parameter and `test_subjects` are distinct,
        but complementary to
        the `subjects` parameter within :class:`Problem_Spec`,
        which is applied after
        selecting the subset of `train_subjects` specified here.

        By default, this value will be set to the special
        str indiciator 'train', which
        specifies that the full set of globally defined training subjects
        (See: :func:`Define_Train_Test_Split`),
        should be used. Other special str indicators
        include 'all' to select all subjects, and 'test'
        to select the test set subjects.

        If `subjects` is passed a str, and that str is not one of
        the str indicators listed above,
        then it will be interpretted as the location of file in which
        to read subjects from (assuming one subjects per line).

        `subjects` may also be a custom array-like of subjects to use.

        Lastly, a special wrapper, Value_Subset, can also be used
        to specify more specific,
        specifically value specific, subsets of subjects to use.
        See :class:`Value_Subset` for how this input wrapper can be used.

        If passing custom input here, be warned that you NEVER
        want to pass an overlap of
        subjects between `train_subjects` and `test_subjects`

        ::

            default = 'train'

    test_subjects : str, array-like or Value_Subset, optional
        This parameter determines the set of testing subjects which are
        used to evaluate the passed instance of :class:`Model_Pipeline`,
        after it
        has been trained on the passed `train_subjects`.

        Note, this parameter and `train_subjects` are distinct,
        but complementary to
        the `subjects` parameter within :class:`Problem_Spec`,
        which is applied after
        selecting the subset of `test_subjects` specified here.

        By default, this value will be set to the special str
        indiciator 'test', which
        specifies that the full set of globally defined
        test subjects
        (See: :func:`Define_Train_Test_Split`),
        should be used. Other special str indicators
        include 'all' to select all subjects, and 'train'
        to select the train set subjects.

        If `subjects` is passed a str, and that str is not one
        of the str indicators listed above,
        then it will be interpretted as the location of file in which
        to read subjects from (assuming one subjects per line).

        `subjects` may also be a custom array-like of subjects to use.

        Lastly, a special wrapper, Value_Subset, can also be used
        to specify more specific,
        specifically value specific, subsets of subjects to use.
        See :class:`Value_Subset` for how this input wrapper can be used.

        If passing custom input here, be warned that you NEVER
        want to pass an overlap of
        subjects between `train_subjects` and `test_subjects`

        ::

            default = 'test'

    feat_importances : :class:`Feat_Importance` list of, str or None, optional

        If passed None, by default, no feature importances will be saved.

        Alternatively, one may pass the keyword 'base', to indicate that
        the base feature importances - those automatically calculated
        by base objects (e.g., beta weights from linear models) be
        saved. In this case, the object Feat_Importance('base') will be made.

        Otherwise, for more detailed control provide here either
        a single, or list of
        :class:`Feat_Importance` param objects
        in which to specify what importance values, and with what
        settings should be computed.
        See the base :class:`Feat_Importance` object for more information
        on how to specify
        these objects.

        See :ref:`Feat Importances` to learn more about feature importances
        generally.

        In this case of a passed list, all passed Feat_Importances
        will attempt to be
        computed.

        ::

            default = None

    return_raw_preds : bool, optional
        If True, return the raw predictions from each fold.

        ::

            default = False

    return_models : bool, optional
        If True, return the trained models from each evaluation.

        ::

            default = False

    run_name : str or 'default', optional
        Each run of test can be optionally
        associated with a specific `run_name`. This name
        is used if `save_results` in
        :func:`Set_Default_ML_Verbosity<BPt_ML.Set_Default_ML_Verbosity>`
        is set to True,
        then will be used as the name output from Test as saved as in the
        specific log_dr
        (if any, and as set when Init'ing
        the :class:`BPt_ML <BPt.BPt_ML>` class object),
        with .test appended to the name.

        If left as 'default', will come up with a kind of
        terrible name passed on the underlying
        model used in the passed `model_pipeline`.

        ::

            default = 'default'

    base_dtype : numpy dtype
        The dataset is cast to a numpy array of float.
        This parameter can be used to change the default
        behavior, e.g., if more resolution or less is needed.

        ::

            default = 'float32'

    Returns
    ----------
    results : dict
        Dictionary containing:
        'scores', the score on the test set by each scorer,
        'raw_preds', A pandas dataframe containing the raw predictions
        for each subject, in the test set, and 'FIs' a list where
        each element corresponds to a passed feature importance.
    '''

    # Perform pre-modeling check
    self._premodel_check()

    # Run initial model pipeline check
    model_pipeline = model_pipeline_check(model_pipeline, self.all_data)

    # Get a free run name
    run_name = get_avaliable_run_name(run_name, model_pipeline)

    # Get the the train subjects + test subjects to use
    _train_subjects = self._get_subjects_to_use(train_subjects)
    _test_subjects = self._get_subjects_to_use(test_subjects)

    # Proc feat importances
    if feat_importances == 'base':
        feat_importances = Feat_Importance(obj='base')
    elif feat_importances == 'default':
        raise RuntimeError('feat_importaces == default is depreciated')

    # Pre-proc problem spec, set as copy ps, right before print
    ps = self._preproc_problem_spec(problem_spec)

    # Run checks before print
    model_pipeline._proc_checks()

    # Print the params being used
    if self.default_ML_verbosity['show_init_params']:

        model_pipeline.print_all(self._print)
        ps.print_all(self._print)

        self._print('Test Params')
        self._print('---------------')
        self._print('train_subjects =', train_subjects)
        self._print('len(train_subjects) =', len(_train_subjects),
                    '(before overlap w/ problem_spec.subjects)')
        self._print('test_subjects =', test_subjects)
        self._print('len(test_subjects) =', len(_test_subjects),
                    '(before overlap w/ problem_spec.subjects)')
        self._print('feat_importances =', feat_importances)
        self._print('run_name =', run_name)
        self._print()

    # Init the Model_Pipeline object with modeling params
    self._init_evaluator(
        model_pipeline=model_pipeline,
        ps=ps,
        cv=None,  # Test doesn't use cv
        feat_importances=feat_importances,
        return_raw_preds=return_raw_preds,
        return_models=return_models,
        base_dtype=base_dtype)

    # Train the model w/ selected parameters and test on test subjects
    train_scores, scores, results =\
        self.evaluator.Test(self.all_data, _train_subjects,
                            _test_subjects)

    results['scores'] = scores

    # Set run name
    if 'FIs' in results:
        for fi in results['FIs']:
            fi.set_target(ps.target)
            fi.set_run_name(run_name)

    # Print out score for all passed scorers
    scorer_strs = self.evaluator.scorer_strs
    self._print()

    score_list, score_type_list = [], []
    if self.default_ML_verbosity['compute_train_score']:
        score_list.append(train_scores)
        score_type_list.append('Training')
        results['train_scores'] = scores

    score_list.append(scores)
    score_type_list.append('Testing')

    for s, name in zip(score_list, score_type_list):

        self._print(name + ' Scores')
        self._print(''.join('_' for i in range(len(name) + 7)))

        for i in range(len(scorer_strs)):

            scorer_name = scorer_strs[i]
            self._print('Scorer: ', scorer_name)

            scr = s[i]
            if len(scr.shape) > 0:

                targets_key = self.evaluator.targets_key

                for score_by_class, class_name in zip(scr, targets_key):
                    self._print('for target class: ', class_name)
                    self._print(name + ' Score: ', score_by_class)
                    self._print()

            else:
                self._print(name + ' Score: ', scr)
                self._print()

    # Add single scores
    for i in range(len(scorer_strs)):
        results[scorer_strs[i]] = scores[i]

        if 'train_scores' in results:
            results['train_' + scorer_strs[i]] = results['train_scores'][i]

    # Save based on default verbosity
    self._save_results(results, run_name + '.test')

    return results


def _premodel_check(self):
    '''Internal helper function
    has been called, and to force a train/test split if not already done.
    Will also call Set_Default_ML_Params if not already called.
    '''

    if self.all_data is None:
        self._print('Calling Prepare_All_Data()',
                    'to change the default merge behavior',
                    'call it again!')
        self.Prepare_All_Data()

    if self.train_subjects is None:

        raise RuntimeError('No train-test set defined!',
                           'If this is intentional, call Train_Test_Split',
                           'with test_size = 0')

    if self.default_ML_verbosity == {}:

        self._print('Setting default ML verbosity settings!')
        self._print('Note, if the following values are not desired,',
                    'call self.Set_Default_ML_Verbosity()')

        self.Set_Default_ML_Verbosity()


def _preproc_param_search(self, object, n_jobs,
                          problem_type,
                          random_state):

    param_search = getattr(object, 'param_search')
    if param_search is None:
        return False

    # Proc param search cv
    if param_search.cv == 'default':
        search_cv = self.cv
    else:
        search_cv =\
            self._get_cv(param_search.cv, show=False)
    param_search.set_cv(search_cv)

    # Set scorer
    param_search.set_scorer(problem_type)

    # Set random_state
    param_search.set_random_state(random_state)

    # Set n_jobs
    param_search.set_n_jobs(n_jobs)

    # Set split vals
    _, split_vals, _ =\
        self._get_split_vals(param_search.splits)
    param_search.set_split_vals(split_vals)

    # Check mp_context for default
    if param_search.mp_context == 'default':
        param_search.mp_context = self.mp_context

    # Set new proc'ed
    setattr(object, 'param_search', param_search)

    return True


def _preproc_cv_splits(self, obj, random_state):

    # Proc cv
    if obj.cv == 'default':
        cv = self.cv
    else:
        cv = self._get_cv(obj.cv, show=False)

    # If CV_Splits
    if isinstance(obj, CV_Splits):

        # Set split vals
        _, split_vals, _ =\
            self._get_split_vals(obj.splits)

        obj.setup(cv=cv, split_vals=split_vals, random_state=random_state)

    # Otherwise must be CV_Split
    else:
        obj.setup(cv=cv, random_state=random_state)


def _preproc_model_pipeline(self, model_pipeline, n_jobs,
                            problem_type, random_state):

    model_pipeline = model_pipeline_check(model_pipeline, self.all_data)

    # Set values across each pipeline pieces params
    model_pipeline.preproc(n_jobs)

    # Pre-proc param search
    has_param_search =\
        self._preproc_param_search(model_pipeline, n_jobs,
                                   problem_type, random_state)

    # If there is a param search set for the Model Pipeline,
    # set n_jobs, the value to pass in the nested check, to 1
    if has_param_search:
        n_jobs = 1

    def nested_model_check(obj):

        # Check for Model or Ensemble
        if isinstance(obj, Model) or isinstance(obj, Ensemble):
            self._preproc_param_search(obj, n_jobs, problem_type, random_state)

        if isinstance(obj, list):
            [nested_model_check(o) for o in obj]
            return

        elif isinstance(obj, dict):
            [nested_model_check(obj[k]) for k in obj]

        elif hasattr(obj, 'get_params'):
            for param in obj.get_params(deep=False):
                nested_model_check(getattr(obj, param))
            return

        return

    # Run nested check
    nested_model_check(model_pipeline)

    def nested_cv_splits_check(obj):

        if isinstance(obj, CV_Splits) or isinstance(obj, CV_Split):
            self._preproc_cv_splits(obj, random_state)

        elif isinstance(obj, list):
            [nested_cv_splits_check(o) for o in obj]

        elif isinstance(obj, dict):
            [nested_cv_splits_check(obj[k]) for k in obj]

        elif hasattr(obj, 'get_params'):
            for param in obj.get_params(deep=False):
                nested_cv_splits_check(getattr(obj, param))

    # Run nested check
    nested_cv_splits_check(model_pipeline)

    return model_pipeline


def _preproc_problem_spec(self, problem_spec):

    # Check if problem_spec is left as default
    if problem_spec == 'default':
        problem_spec = Problem_Spec()

    # Set ps to copy of problem spec and init
    ps = deepcopy(problem_spec)

    # Update target with actual target key
    target_key = self._get_targets_key(ps.target)
    ps.set_params(target=target_key)

    # Replace problem_spec type w/ correct if passed short hands
    pt = ps.problem_type
    if pt == 'default':

        # For future multi-label support...
        if isinstance(target_key, list):
            pt = 'multilabel'
        else:
            pt = get_target_type(self.all_data[target_key])

        self._print('problem_spec problem_type ==  default, setting as:',
                    pt)

    elif pt == 'b':
        pt = 'binary'

    elif pt == 'c' or pt == 'multiclass':
        pt = 'categorical'

    elif pt == 'f' or pt == 'float':
        pt = 'regression'

    ps.problem_type = pt

    # Check for if default scorer
    if ps.scorer == 'default':
        default_scorers = {'regression': ['explained_variance',
                                          'neg_mean_squared_error'],
                           'binary': ['matthews', 'roc_auc',
                                      'balanced_accuracy'],
                           'categorical': ['matthews', 'roc_auc_ovr',
                                           'balanced_accuracy']}
        ps.scorer = default_scorers[pt]
        self._print('problem_spec scorer ==  default, setting as:', ps.scorer)

    # Proc subjects to use
    final_subjects = self._get_subjects_to_use(ps.subjects)
    ps.set_final_subjects(final_subjects)

    # Set by class defaults
    if ps.n_jobs == 'default':
        ps.n_jobs = self.n_jobs

    if ps.random_state == 'default':
        ps.random_state = self.random_state

    # If any input has changed, manually (i.e., not by problem_spec init)
    ps._proc_checks()

    return ps


def _get_split_vals(self, splits):

    if isinstance(splits, int) or isinstance(splits, float):
        split_names, split_vals, sv_le = None, None, None

    else:
        split_names = self._add_strat_u_name(splits)

        if isinstance(split_names, str):
            split_names = [split_names]

        split_vals, sv_le =\
            get_unique_combo_df(self.strat, split_names)

    return split_names, split_vals, sv_le


def _get_subjects_to_use(self, subjects_to_use):

    # If str passed, either loc to load, or train, test, all
    if isinstance(subjects_to_use, str):
        if subjects_to_use == 'all':
            subjects = self.all_data.index
        elif subjects_to_use == 'train':
            subjects = self.train_subjects
        elif subjects_to_use == 'test':
            subjects = self.test_subjects
        else:
            subjects = self._load_set_of_subjects(loc=subjects_to_use)

    # Other case is if passed value subset, determine the values to use
    elif is_value_subset(subjects_to_use):

        split_names, split_vals, sv_le =\
            self._get_split_vals(subjects_to_use.name)

        selected = split_vals[split_vals == subjects_to_use.value]
        subjects = set(selected.index)

        rev_values = reverse_unique_combo_df(selected, sv_le)[0]

        last_subjects_names = []
        for strat_name, value in zip(split_names, rev_values):
            if self.strat_u_name in strat_name:
                strat_name = strat_name.replace(self.strat_u_name, '')

            last_subjects_names.append((strat_name, value))

        self._print('subjects set to: ',
                    last_subjects_names)
        self._print()

    # Can also be values subset
    elif is_values_subset(subjects_to_use):

        # Add strat u name to name if not already added
        name = self._add_strat_u_name(subjects_to_use.name)

        # Extract the values as list
        values = conv_to_list(subjects_to_use.values)

        # Check name to make sure loaded
        if name not in self.all_data:
            raise ValueError(name, 'is not a valid loaded Strat feature!')

        # Get by value
        subjects = self.all_data[self.all_data[name].isin(values)].index

        # Make sure subjects is set-like
        subjects = set(list(subjects))

    # Lastly, if not the above, assume it is an array-like of subjects
    else:
        subjects = self._load_set_of_subjects(subjects=subjects_to_use)

    return subjects


def get_pipeline(self, model_pipeline, problem_spec,
                 progress_loc=None, has_search=False):

    # If has search is False, means this is the top level
    # or the top level didnt have a search
    if not has_search:
        if model_pipeline.param_search is None:
            has_search = False
        else:
            has_search = True

    # If either this set of model_pipeline params or the parent
    # params had search params, then a copy of problem spec with n_jobs set
    # to 1 should be passed to children get pipelines,
    if has_search:
        nested_ps = copy.deepcopy(problem_spec)
        nested_ps.n_jobs = 1
    else:
        nested_ps = problem_spec

    def nested_check(obj):

        if hasattr(obj, 'obj') and isinstance(obj.obj, Model_Pipeline):

            nested_pipe, nested_pipe_params =\
                self.get_pipeline(model_pipeline=obj.obj,
                                  problem_spec=nested_ps,
                                  progress_loc=progress_loc,
                                  has_search=has_search)

            # Set obj as nested pipeline
            setattr(obj, 'obj', nested_pipe)

            # Set obj's params as the nested_pipe_params
            setattr(obj, 'params', nested_pipe_params)
            return

        if isinstance(obj, list):
            [nested_check(o) for o in obj]
            return

        elif isinstance(obj, dict):
            [nested_check(obj[k]) for k in obj]
            return

        if hasattr(obj, 'get_params'):
            for param in obj.get_params(deep=False):
                nested_check(getattr(obj, param))
            return

    # Run nested check on model_pipeline
    nested_check(model_pipeline)

    # Preproc model
    model_pipeline = self._preproc_model_pipeline(model_pipeline,
                                                  problem_spec.n_jobs,
                                                  problem_spec.problem_type,
                                                  problem_spec.random_state)

    # Init data scopes
    self.Data_Scopes.set_all_keys(problem_spec)

    # Return the pipeline
    return get_pipe(pipeline_params=model_pipeline,
                    problem_spec=problem_spec,
                    Data_Scopes=self.Data_Scopes,
                    progress_loc=progress_loc,
                    verbose=self.default_ML_verbosity['pipeline_verbose'])


def _init_evaluator(self, model_pipeline, ps,
                    cv, feat_importances, return_raw_preds,
                    return_models, base_dtype):

    # Make copies of the passed pipeline
    # and only make changes and pass along the copies
    pipe = copy.deepcopy(model_pipeline)

    # Calling get pipeline performs preproc on model_pipeline
    # and Data_Scopes
    model, _ =\
        self.get_pipeline(
            pipe, ps, progress_loc=self.default_ML_verbosity['progress_loc'])

    # Set the evaluator obj
    self.evaluator =\
        Evaluator(model=model,
                  problem_spec=ps,
                  cv=cv,
                  all_keys=self.Data_Scopes.all_keys,
                  feat_importances=feat_importances,
                  return_raw_preds=return_raw_preds,
                  return_models=return_models,
                  verbosity=self.default_ML_verbosity,
                  base_dtype=base_dtype,
                  _print=self._ML_print)


def _handle_scores(self, scores, name, weight_scorer, n_repeats, run_name,
                   n_splits, summary_dfs):

    all_summary_scores = []
    summary_dfs[name] = pd.DataFrame()

    scorer_strs = self.evaluator.scorer_strs

    self._print(name + ' Scores')
    self._print(''.join('_' for i in range(len(name) + 7)))

    weight_scorers = conv_to_list(weight_scorer, len(scorer_strs))

    for i in range(len(scorer_strs)):

        # Weight outputed scores if requested
        if weight_scorers[i]:
            weights = self.evaluator.n_test_per_fold
        else:
            weights = None

        scorer_name = scorer_strs[i]
        self._print('Scorer: ', scorer_name)
        score_by_scorer = scores[:, i]

        if len(score_by_scorer[0].shape) > 0:
            by_class = [[score_by_scorer[i][j] for i in
                        range(len(score_by_scorer))] for j in
                        range(len(score_by_scorer[0]))]

            summary_scores_by_class =\
                [compute_micro_macro(class_scores, n_repeats,
                 n_splits, weights=weights) for class_scores in by_class]

            targets_key = self.evaluator.targets_key
            classes = self.evaluator.classes

            class_names =\
                self.targets_encoders[targets_key].inverse_transform(
                    classes.astype(int))

            for summary_scores, class_name in zip(summary_scores_by_class,
                                                  class_names):

                self._print('Target class: ', class_name)
                self._print_summary_score(name, summary_scores,
                                          n_repeats, run_name,
                                          scorer_name, class_name,
                                          weights=weights)

            all_summary_scores.append(summary_scores_by_class)

        else:

            # Compute macro / micro summary of scores
            summary_scores = compute_micro_macro(score_by_scorer,
                                                 n_repeats,
                                                 n_splits,
                                                 weights=weights)

            self._print_summary_score(name, summary_scores,
                                      n_repeats, run_name,
                                      scorer_name, weights=weights)

            all_summary_scores.append(summary_scores)

    return all_summary_scores


def _print_summary_score(self, name, summary_scores, n_repeats, run_name,
                         scorer_name, class_name=None, weights=None):

    mn = 'Mean'
    if weights is not None:
        mn = 'Weighted ' + mn

    self._print(mn + ' ' + name + ' score: ', summary_scores[0])

    if n_repeats > 1:
        self._print('Micro Std in ' + name + ' score: ',
                    summary_scores[1])
        self._print('Macro Std in ' + name + ' score: ',
                    summary_scores[2])
    else:
        self._print('Std in ' + name + ' score: ',
                    summary_scores[1])

    self._print()


def _save_results(self, results, save_name):

    if self.default_ML_verbosity['save_results'] and self.log_dr is not None:

        save_dr = os.path.join(self.exp_log_dr, 'results')
        os.makedirs(save_dr, exist_ok=True)

        save_spot = os.path.join(save_dr, save_name)

        append, cnt = '', 0
        while os.path.exists(save_spot+append):
            cnt += 1
            append = str(cnt)

        with open(save_spot+append, 'wb') as f:
            pkl.dump(results, f)


def model_pipeline_check(model_pipeline, data):

    # Add checks on Model_Pipeline
    if not isinstance(model_pipeline, Model_Pipeline):

        # Check for if model str first
        if isinstance(model_pipeline, str):
            model_pipeline = Model(obj=model_pipeline)

        # In case of passed valid single model, wrap in Model_Pipeline
        if hasattr(model_pipeline, '_is_model'):
            model_pipeline = Model_Pipeline(imputers=None,
                                            model=model_pipeline)
        else:
            raise RuntimeError('model_pipeline must be a Model_Pipeline',
                               ' model str or Model-like')

    # Early check to see if imputer is needed
    model_pipeline.check_imputer(data)

    return model_pipeline



import pandas as pd
import numpy as np
from itertools import combinations


def proc_fop(fop):

    # If provided as just % number, divide by 100
    if not isinstance(fop, tuple):
        fop /= 100
        return (fop, 1-fop)

    elif fop[0] is None:
        return tuple([None, 1-(fop[1] / 100)])

    elif fop[1] is None:
        return tuple([fop[0] / 100, None])

    return tuple([fop[0]/100, 1-(fop[1] / 100)])


def load_subjects(subjects):

    # Include checks here for Value_Subset and Value Subsets!!!
    # @TODO

    loaded_subjects = set()

    if isinstance(subjects, str):
        with open(subjects, 'r') as f:
            lines = f.readlines()

            for line in lines:
                subject = line.rstrip()
                loaded_subjects.add(subject)

    else:
        loaded_subjects = set([s for s in subjects])

    return loaded_subjects


class Dataset(pd.DataFrame):

    ROLES = set(['data', 'target', 'non input'])
    RESERVED_SCOPES = set(['all', 'float', 'category',
                           'data', 'data file',
                           'non input', 'target'])
    _metadata = ['roles', 'scopes', 'name_map']

    @property
    def _constructor(self):
        return Dataset

    @property
    def _constructor_sliced(self):
        return pd.Series

    def _check_name_map(self):

        if not hasattr(self, 'name_map'):
            self.name_map = {}

    def apply_inclusions(self, subjects):

        # Load inclusions
        inclusions = load_subjects(subjects)

        if len(inclusions) > 0:
            to_drop = set(self.index) - inclusions
            self.drop(to_drop, axis=0, inplace=True)

        return self

    def apply_exclusions(self, subjects):

        # Load exclusions
        exclusions = load_subjects(subjects)

        if len(exclusions) > 0:
            to_drop = set(self.index).intersection(exclusions)
            self.drop(to_drop, axis=0, inplace=True)

        return self

    def _check_roles(self):

        if not hasattr(self, 'roles'):
            self.roles = {}

        # Fill in any column without a role.
        for col in list(self.columns):
            if col not in self.roles:
                self._set_role(col, 'data')

        # Remove any saved roles that are not
        # in the data any more.
        for col in list(self.roles):
            if col not in self.columns:
                del self.roles[col]

    def get_roles(self):

        self._check_roles()
        return self.roles

    def set_role(self, col, role):

        self._check_roles()
        self._set_role(col, role)

        return self

    def set_roles(self, col_to_roles):

        self._check_roles()
        for col, role in zip(col_to_roles, col_to_roles.values()):
            self._set_role(col, role)

        return self

    def _set_role(self, col, role):

        if role not in self.ROLES:
            raise AttributeError(
                'Passed role "' + str(role) + '" must be one of ' +
                str(self.ROLES))

        # Set as role
        self.roles[col] = role

        return self

    def _check_scopes(self):

        # If doesnt exist, create scopes
        if not hasattr(self, 'scopes'):
            self.scopes = {}

        # Make sure each col is init'ed with a scope
        for col in self.columns:
            if col not in self.scopes:
                self.scopes[col] = set()

        # Make sure to get rid of any removed cols
        for col in list(self.scopes):
            if col not in self.columns:
                del self.scopes[col]

        # Make sure scopes includes if categorical or not
        for col in self.columns:
            if self[col].dtype.name == 'category':
                self._add_scope(col, 'category')
            else:
                self._remove_scope(col, 'category')

        return self

    def get_scopes(self):

        self._check_scopes()
        return self.scopes

    def add_scope(self, col, scope):

        self._check_scopes()
        self._add_scope(col, scope)

        return self

    def add_scopes(self, col_to_scopes):

        self._check_scopes()
        for col, scope in zip(col_to_scopes, col_to_scopes.values()):
            self._add_scope(col, scope)

        return self

    def _add_scope(self, col, scope):

        if scope == 'category':
            self.scopes[col].add(scope)
            self[col] = self[col].astype('category')

        elif isinstance(scope, str):
            self.scopes[col].add(scope)

        elif isinstance(scope, int) or isinstance(scope, float):
            self.scopes[col].add(str(scope))

        else:
            for s in set(list(scope)):
                self._add_scope(col, s)

        return self

    def remove_scope(self, col, scope):

        self._check_scopes()
        self._remove_scope(col, scope)

        return self

    def remove_scopes(self, col_to_scopes):

        self._check_scopes()
        for col, scope in zip(col_to_scopes, col_to_scopes.values()):
            self._remove_scope(col, scope)

        return self

    def _remove_scope(self, col, scope):

        try:
            self.scopes[col].remove(scope)
        except KeyError:
            pass

        return self

    def _get_cols_from_scope(self, scope):

        # Perform checks first
        self._check_scopes()
        self._check_roles()

        saved_scopes = set()
        for col in self.columns:
            saved_scopes.update(self.scopes[col])

        # Check is passed scope is reserved
        if isinstance(scope, str):
            if scope in self.RESERVED_SCOPES:

                if scope == 'all':
                    return list(self.columns)

                elif scope == 'float':
                    return [col for col in self.columns if
                            'category' not in self.scopes[col] and
                            'data file' not in self.scopes[col]]

                elif scope == 'category':
                    return [col for col in self.columns if
                            'category' in self.scopes[col]]

                elif scope == 'data':
                    return [col for col in self.columns if
                            self.roles[col] == 'data']

                elif scope == 'target':
                    return [col for col in self.columns if
                            self.roles[col] == 'target']

                elif scope == 'non input':
                    return [col for col in self.columns if
                            self.roles[col] == 'non input']

                elif scope == 'data file':
                    return [col for col in self.columns if
                            'data file' in self.scopes[col]]

            # Check if passed scope is a loaded column
            elif scope in self.columns:
                return [scope]

            # Check if a saved scope
            elif scope in saved_scopes:
                return [col for col in self.columns if
                        scope in self.scopes[col]]

            # Do a search, see if passed scope is a stub of any strings
            else:
                return [col for col in self.columns if
                        scope in col]

        cols = []
        for scp in scope:
            cols += self._get_cols_from_scope(scp)

        return sorted(list(set(cols)))

    def auto_detect_categorical(self):
        pass

    def filter_outliers_by_percent(self, fop=1, scope='float', drop=True):

        # Proc fop
        fop = proc_fop(fop)
        lower, upper = fop

        # Get cols from scope
        cols = self._get_cols_from_scope(scope)

        if drop:
            to_drop = set()

            if lower is not None:
                mask = self[cols] < self[cols].quantile(lower)
                to_drop.update(set(self[cols][mask.any(axis=1)].index))

            if upper is not None:
                mask = self[cols] > self[cols].quantile(upper)
                to_drop.update(set(self[cols][mask.any(axis=1)].index))

            self.drop(list(to_drop), inplace=True)

        else:

            for c in cols:
                if lower is not None:
                    l_thresh = self[c] < self[c].quantile(lower)
                else:
                    l_thresh = None

                if upper is not None:
                    u_thresh = self[c] > self[c].quantile(upper)
                else:
                    u_thresh = None

                if lower is not None:
                    self.loc[l_thresh, c] = np.nan
                if upper is not None:
                    self.loc[u_thresh, c] = np.nan

        return self

    def filter_outliers_by_std(self, n_std=10, scope='float', drop=True):

        if not isinstance(n_std, tuple):
            n_std = (n_std, n_std)

        # Get cols from scope
        cols = self._get_cols_from_scope(scope)

        if drop:

            mean = self[cols].mean()
            std = self[cols].std()

            to_drop = set()

            if n_std[0] is not None:
                l_scale = mean - (n_std[0] * std)
                mask = self[cols] < l_scale
                to_drop.update(set(self[cols][mask.any(axis=1)].index))

            if n_std[1] is not None:
                u_scale = mean + (n_std[1] * std)
                mask = self[cols] > u_scale
                to_drop.update(set(self[cols][mask.any(axis=1)].index))

            self.drop(list(to_drop), inplace=True)

        else:

            for c in cols:

                if n_std[0] is not None:
                    l_scale = self[c].mean() - (n_std[0] * self[c].std())
                else:
                    l_scale = None

                if n_std[1] is not None:
                    u_scale = self[c].mean() + (n_std[1] * self[c].std())
                else:
                    u_scale = None

                if n_std[0] is not None:
                    self.loc[self[c] < l_scale, c] = np.nan
                if n_std[1] is not None:
                    self.loc[self[c] > u_scale, c] = np.nan

        return self

    def filter_categorical_by_percent(self, scope):

        return self

        # Add cast to categorical
        pass

    def drop_non_unique(self, scope='all'):
        '''This method will drop any columns with only one unique value.'''

        # Get cols from scope
        cols = self._get_cols_from_scope(scope)

        to_drop = []
        for col in cols:
            if len(self[col].unique()) == 1:
                to_drop.append(col)

        self.drop(to_drop, axis=1, inplace=True)
        return self

    def drop_id_cols(self, scope='all'):
        '''This method will drop any str-type / object type columns
        where the number of unique columns is equal
        to the length of the dataframe.'''

        # Get cols from scope
        cols = self._get_cols_from_scope(scope)

        to_drop = []

        for col in cols:
            if self[col].dtype.name == 'object':
                if len(self[col].unique()) == len(self):
                    to_drop.append(col)

        self.drop(to_drop, axis=1, inplace=True)
        return self

    def drop_duplicate_cols(self, scope='all'):

        # Get cols from scope
        cols = self._get_cols_from_scope(scope)

        to_drop = []
        for col1, col2 in combinations(cols, 2):
            if np.array_equal(self[col1], self[col2]):
                to_drop.append(col2)

        self.drop(to_drop, axis=1, inplace=True)
        return self

    def drop_cols(self,
                  exclusions=None,
                  inclusions=None,
                  scope='all'):

        # Get cols from scope
        cols = self._get_cols_from_scope(scope)

        if exclusions is not None:
            exclusions = self._get_cols_from_scope(exclusions)
        else:
            exclusions = []

        if inclusions is not None:
            inclusions = self._get_cols_from_scope(inclusions)
        else:
            inclusions = cols

        to_drop = []
        for col in cols:
            if col in exclusions or col not in inclusions:
                to_drop.append(col)

        self.drop(to_drop, axis=1, inplace=True)
        return self

    def _add_new_copy(self, old, new):

        # Copy as new key
        self[new] = self[old].copy()

        # Try to copy scope and role too
        try:
            self.scopes[new] = self.scopes[old].copy()
        except KeyError:
            pass

        try:
            self.roles[new] = self.roles[old]
        except KeyError:
            pass

    def binarize(self, scope, threshold=None, lower=None,
                 upper=None, replace=True):

        if threshold is None and lower is None and upper is None:
            raise RuntimeError('Some value must be set.')
        if lower is not None and upper is None:
            raise RuntimeError('Upper must be set.')
        if upper is not None and lower is None:
            raise RuntimeError('Lower must be set.')

        # Make sure name_map init'ed
        self._check_name_map()

        # Get cols from scope
        cols = self._get_cols_from_scope(scope)

        # Binarize each column
        for col in cols:
            self._binarize(col, threshold, lower, upper, replace)

        return self

    def _binarize(self, col, threshold, lower, upper, replace):

        # Extract values / series
        values = self[col]

        if not replace:
            new_key = col + '_binary'

            if new_key in self.columns:
                cnt = 1
                while new_key + str(cnt) in self.columns:
                    cnt += 1

                new_key += str(cnt)

            # Make a copy under the new key
            self._add_new_copy(old=col, new=new_key)

            # Change col to new_key
            col = new_key

        # If upper and lower passed instead of threshold
        if threshold is None:

            # Drop between the values
            to_drop = values[(values <= upper) &
                             (values >= lower)].index
            self.drop(to_drop, axis=0, inplace=True)

            # Binarize
            self[col] = self[col].where(self[col] > lower, 0)
            self[col] = self[col].where(self[col] < upper, 1)

            # Add to name map
            self.name_map[col] =\
                {0: '<' + str(lower), 1: '>' + str(upper)}

        # If a single threshold
        else:

            # Binarize
            self[col] = self[col].where(self[col] >= threshold, 0)
            self[col] = self[col].where(self[col] < threshold, 1)

            # Add to name map
            self.name_map[col] =\
                {0: '<' + str(threshold), 1: '>=' + str(threshold)}

        # Make sure category scope
        self.add_scope(col, 'category')

    def k_bin(self, scope):
        pass

    def ordinalize(self, scope):
        pass

    def nan_to_class(self, scope):

        # Have checks s.t., only works if all columns are categorical
        pass

    def drop_by_unique_val(self, scope):
        pass

    def print_nan_info(self, scope):
        pass

    # need to handle data files still***
    def add_data_files(self):
        pass

    def rename(self, **kwargs):
        print('Warning: Renaming might cause errors!')
        super().rename(**kwargs)

    from .Dataset_Plotting import (plot,
                                   _plot_category)


def get_fake_dataset():

    fake = Dataset()
    fake['1'] = [1, 2, 3]
    fake['2'] = ['6', '7', '8']
    fake['2'] = fake['2'].astype('category')
    fake['3'] = [np.nan, 2, 3]

    return fake


def get_fake_dataset2():

    fake = Dataset()
    fake['1'] = [1, 1, 1]
    fake['2'] = [1, 2, 3]
    fake['3'] = ['1', '2', '3']

    return fake


def get_fake_dataset3():

    fake = Dataset()
    fake['1'] = [1, 1, 1]
    fake['2'] = [1, 1, 1]
    fake['3'] = ['2', '2', '2']
    fake['4'] = ['2', '2', '2']
    fake['5'] = ['2', 1, '2']

    return fake


def test_add_scope():

    df = get_fake_dataset()

    df.add_scope(col='1', scope='a')
    assert df.scopes['1'] == set(['a'])

    df.add_scope(col='1', scope='b')
    assert df.scopes['1'] != set(['a'])
    assert df.scopes['1'] == set(['a', 'b'])

    # Test some get cols from scope
    assert(set(df._get_cols_from_scope('a')) == set(['1']))
    assert(set(df._get_cols_from_scope('b')) == set(['1']))
    assert(set(df._get_cols_from_scope(['a', 'b'])) == set(['1']))

    df = get_fake_dataset()
    df.add_scope(col='1', scope='category')
    assert(df['1'].dtype.name == 'category')


def test_set_roles():

    df = get_fake_dataset()
    df.set_role('1', 'target')
    df.set_role('2', 'non input')

    assert(set(df._get_cols_from_scope('target')) == set(['1']))
    assert(set(df._get_cols_from_scope('non input')) == set(['2']))


def test_get_cols_from_scope():

    df = get_fake_dataset()
    assert(set(df._get_cols_from_scope('all')) == set(['1', '2', '3']))
    assert(set(df._get_cols_from_scope('data')) == set(['1', '2', '3']))
    assert(set(df._get_cols_from_scope('1')) == set(['1']))
    assert(set(df._get_cols_from_scope('category')) == set(['2']))


def test_filter_outliers():

    df = get_fake_dataset()
    df.filter_outliers_by_percent(20, scope='3', drop=False)
    assert pd.isnull(df['3']).all()


def test_drop_non_unique():

    df = get_fake_dataset2()
    df.drop_non_unique()

    assert '1' not in df
    assert '2' in df
    assert '3' in df


def test_drop_id_cols():

    df = get_fake_dataset2()
    df.drop_id_cols()

    assert '1' in df
    assert '2' in df
    assert '3' not in df


def test_drop_duplicate_cols():

    df = get_fake_dataset3()
    df.drop_duplicate_cols()

    assert '5' in df
    assert df.shape == (3, 3)


def test_apply_inclusions():

    df = get_fake_dataset3()
    df.apply_inclusions([0])
    assert len(df) == 1


def test_apply_exclusions():

    df = get_fake_dataset()
    df.apply_exclusions([0, 1])
    assert len(df) == 1

    df = get_fake_dataset()
    df.apply_exclusions([0])
    assert len(df) == 2


def test_drop_cols_inclusions():

    df = get_fake_dataset()
    df.drop_cols(inclusions='1')
    assert '1' in df
    assert df.shape[1] == 1

    df = get_fake_dataset()
    df.drop_cols(inclusions='category')
    assert '2' in df

    df = get_fake_dataset()
    df.drop_cols(inclusions=['1', '2'])
    assert df.shape[1] == 2

    df = Dataset(columns=['xxx1', 'xxx2', 'xxx3', '4'])
    df.drop_cols(inclusions=['xxx'])
    assert '4' not in df
    assert df.shape[1] == 3


def test_drop_cols_exclusions():

    df = get_fake_dataset()
    df.drop_cols(exclusions='1')
    assert '1' not in df
    assert df.shape[1] == 2

    df = get_fake_dataset()
    df.drop_cols(exclusions=['1', '2'])
    assert '3' in df
    assert df.shape[1] == 1

    df = Dataset(columns=['xxx1', 'xxx2', 'xxx3', '4'])
    df.drop_cols(exclusions=['xxx'])
    assert '4' in df
    assert df.shape[1] == 1


def test_binarize_threshold():

    df = get_fake_dataset()
    df.binarize('1', threshold=1.5)

    assert df.loc[0, '1'] == 0
    assert df.loc[1, '1'] == 1
    assert 'category' in df.scopes['1']
    assert df.name_map['1'] == {0: '<1.5', 1: '>=1.5'}


def test_binarize_upper_lower():

    df = get_fake_dataset()
    df.binarize('1', lower=2, upper=2)

    assert len(df) == 2
    assert df.loc[0, '1'] == 0
    assert df.loc[2, '1'] == 1
    assert 'category' in df.scopes['1']
    assert df.name_map['1'] == {0: '<2', 1: '>2'}


test_add_scope()
test_set_roles()
test_get_cols_from_scope()
test_filter_outliers()
test_drop_non_unique()
test_drop_id_cols()
test_drop_duplicate_cols()
test_apply_inclusions()
test_apply_exclusions()
test_drop_cols_inclusions()
test_drop_cols_exclusions()
test_binarize_threshold()
test_binarize_upper_lower()



"""
Data_Helpers.py
====================================
Various helper functions for loading and processing data for BPt.
Specifically, these are non-class functions used in _Data.py and BPt_ML.py.
"""
import numpy as np
import numpy.ma as ma
import random
from sklearn.preprocessing import KBinsDiscretizer
from sklearn.preprocessing import LabelEncoder
from operator import add
from functools import reduce
import warnings
import pandas as pd
from ..helpers.Data_File import load_data_file_proxies


def auto_data_type(col):
    '''Assumes passed non nan data'''

    # First check for binary by unique number of columns
    n_unique = len(col.unique())
    if n_unique == 2:
        return 'binary'

    # If loaded as category data-type
    if col.dtype.name == 'category':
        return 'categorical'

    # If object dtype and less then 30 unique categories
    elif col.dtype.name == 'object' and n_unique < 30:
        return 'categorical'

    # All else, float
    return 'float'


def get_non_drop(data, key, drop_val):

    # If drop_val np.nan, means treating dropped as missing value
    if drop_val is np.nan:
        non_drop_subjects = data[~(data[key].isna())].index

    # Otherwise
    else:
        non_drop_subjects = data[~(data[key] == drop_val)].index

    non_drop_data = data.loc[non_drop_subjects]
    return non_drop_data, non_drop_subjects


def fill_df_with(data, subjects, fill_data):

    for key in list(data):

        dtype = fill_data[key].dtype.name
        o_dtype = data[key].dtype.name

        data.loc[subjects, key] =\
            fill_data.loc[subjects, key].astype(o_dtype)
        data[key] = data[key].astype(dtype)

    return data


def process_binary_input(data, key, drop_val=np.nan,
                         _print=print):
    '''Helper function to perform processing on binary input

    Parameters
    ----------
    data : pandas DataFrame
        BPt formatted df. Must contain a column with `key`

    key : str
        Column key of the column to process within `data` input.

    drop_val : NaN or int
        If a row needs to be dropped, replace with drop_val

        (default = np.nan)

    _print : print func, optional
        Either python print statement or overriden print
        func.

        (default = print)

    Returns
    ----------
    pandas DataFrame
        The post-processed BPt formatted input df.

    sklearn LabelEncoder
        The sklearn labelencoder object mapping input to
        transformed binary label.
    '''

    unique_vals, counts = np.unique(data[key], return_counts=True)

    if len(unique_vals) == 1:
        warnings.warn('Binary input ' + repr(key) + ' was '
                      'passed with only 1 unique value')

    # Preform check for mistaken values
    # Assuming should be binary, so 2 unique values
    if len(unique_vals) > 2:

        # Select top two scores by count
        keep_inds = np.argpartition(counts, -2)[-2:]
        keep_vals = unique_vals[keep_inds]
        keep_vals.sort()

        to_drop = data.index[~data[key].isin(keep_vals)]
        data.loc[to_drop, key] = drop_val

        _print('More than two unique score values found,',
               'filtered all but', keep_vals)

    # Work on only non_dropped data / non NaN data if applicable
    non_drop_data, non_drop_subjects =\
        get_non_drop(data, key, drop_val)

    # Perform actual binary encoding
    encoder = LabelEncoder()
    non_drop_data[key] = encoder.fit_transform(np.array(non_drop_data[key]))
    non_drop_data[key] = non_drop_data[key].astype('category')

    if len(np.unique(non_drop_data[key])) > 2:
        raise ValueError('Binary type, but more than two unique values found '
                         'for input' + repr(key))

    data = fill_df_with(data, non_drop_subjects, non_drop_data)

    return data, encoder


def process_ordinal_input(data, key, drop_percent=None,
                          drop_val=np.nan, nac=False, _print=print):
    '''Helper function to perform processing on ordinal input,
    where note this definition of ordinal means categorical ordinal...
    so converting input to ordinal, but from categorical!

    Parameters
    ----------
    data : pandas DataFrame
        BPt formatted df. Must contain a column with `key`

    key : str
        Column key of the column to process within `data` input.

    drop_percent : float
        % to drop

    drop_val : NaN or int
        If a row needs to be dropped, replace with drop_val

        (default = np.nan)

    Returns
    ----------
    pandas DataFrame
        The post-processed BPt formatted input df.

    sklearn LabelEncoder
        The sklearn labelencoder object mapping input to
        transformed ordinal label
    '''

    # If requested, perform categorical dropping by percent
    if drop_percent:
        data = cat_drop_by_percent(data, key, drop_percent, drop_val, _print)

    # Work on only non_dropped data / non NaN data if applicable
    non_drop_data, non_drop_subjects =\
        get_non_drop(data, key, drop_val)

    # Encode ordinally
    label_encoder = LabelEncoder()
    non_drop_data[key] = label_encoder.fit_transform(non_drop_data[key])
    non_drop_data[key] = non_drop_data[key].astype('category')

    # Re-create data
    data = fill_df_with(data, non_drop_subjects, non_drop_data)

    # Check if nan was encoded if nac
    if nac:
        try:
            nan_val = label_encoder.transform(['nan'])[0]
            label_encoder.nan_val = nan_val

        # Still need a NaN value, even if not already set
        except ValueError:

            label_encoder.classes_ =\
                np.array(list(label_encoder.classes_) + ['nan'])
            label_encoder.nan_val = len(label_encoder.classes_) - 1

    return data, label_encoder


def process_float_input(data, key, bins, strategy,
                        drop_percent=None, drop_val=np.nan,
                        nac=False, _print=print):

    # Data will come in either as w/o NaN's or everything as strs
    non_nan_subjects = data[data[key].astype('str') != 'nan'].index
    non_nan_data = data.loc[non_nan_subjects]

    encoder = KBinsDiscretizer(n_bins=bins, encode='ordinal',
                               strategy=strategy)

    vals = np.array(non_nan_data[key].astype('float')).reshape(-1, 1)
    vals = np.squeeze(encoder.fit_transform(vals))
    _print('KBins encoded', encoder.n_bins, 'bins')

    # Put back in data
    data.loc[non_nan_subjects, key] = vals

    # Set nan val if nan as class
    if nac:
        nan_val = np.max(vals) + 1

        # If any NaNs replace with
        nan_subjects = data[data[key].astype('str') == 'nan'].index
        if len(nan_subjects) > 0:
            data.loc[nan_subjects, key] = nan_val

        # Save nan val in the encoder
        encoder.nan_val = nan_val

    # If requested, perform categorical dropping by percent
    if drop_percent:
        data = cat_drop_by_percent(data, key, drop_percent, drop_val, _print)

    # Set type to category
    data[key] = data[key].astype('category')

    return data, encoder


def cat_drop_by_percent(data, key, drop_percent, drop_val, _print):

    unique_vals, counts = np.unique(data[key], return_counts=True)

    drop_inds = np.where(counts / len(data) < drop_percent)
    drop_vals = unique_vals[drop_inds]

    to_drop = data.index[data[key].isin(drop_vals)]
    data.loc[to_drop, key] = drop_val

    _print('Dropped value(s)', drop_vals, 'according to passed drop',
           'percent of', drop_percent)

    return data


def get_unused_drop_val(data):

    drop_val = random.random()
    while (data == drop_val).any().any():
        drop_val = random.random()
    return drop_val


def proc_fop(fop):

    # If provided as just % number, divide by 100
    if not isinstance(fop, tuple):
        fop /= 100
        return (fop, 1-fop)

    elif fop[0] is None:
        return tuple([None, 1-(fop[1] / 100)])

    elif fop[1] is None:
        return tuple([fop[0] / 100, None])

    return tuple([fop[0]/100, 1-(fop[1] / 100)])


def filter_float_by_outlier(data, key, filter_outlier_percent,
                            drop_val=999, _print=print):
    '''Helper function to perform filtering on a dataframe,
    by setting values to be NaN, then optionally removing rows inplace

    Parameters
    ----------
    data : pandas DataFrame
        BPt formatted df. Must contain a column with `key`

    key : str
        Column key of the column to process within `data` input.

    filter_outlier_percent : int, float, tuple or None
        A percent of values to exclude from either end of the
        score distribution, provided as either 1 number,
        or a tuple (% from lower, % from higher).
        set `filter_outlier_percent` to None for no filtering.
        If over 1, then treated as a percent.

    _print : print func, optional
        Either python print statement or overriden print
        func.

        (default = print)

    Returns
    ----------
    pandas DataFrame
        The post-processed BPt formatted input df.
    '''

    # For length of code / readability
    fop = proc_fop(filter_outlier_percent)

    _print('Filtering for outliers, dropping rows with params: ', fop)
    _print('Min-Max value (pre-filter):',
           np.nanmin(data[key]), np.nanmax(data[key]))

    if fop[0] is not None:
        q1 = data[key].quantile(fop[0])
        data.loc[data[key] < q1, key] = drop_val
    if fop[1] is not None:
        q2 = data[key].quantile(fop[1])
        data.loc[data[key] > q2, key] = drop_val

    _print('Min-Max value (post outlier filtering):',
           np.nanmin(data[key][data[key] != drop_val]),
           np.nanmax(data[key][data[key] != drop_val]))

    return data


def filter_float_by_std(data, key, n_std,
                        drop_val=999, _print=print):

    if not isinstance(n_std, tuple):
        n_std = (n_std, n_std)

    _print('Filtering for outliers by stds:', n_std)
    _print('Min-Max value (pre-filter):',
           np.nanmin(data[key]), np.nanmax(data[key]))

    mean = data[key].mean()
    std = data[key].std()

    if n_std[0] is not None:
        l_scale = n_std[0] * std
        data.loc[data[key] < mean - l_scale, key] = drop_val

    if n_std[1] is not None:
        u_scale = n_std[1] * std
        data.loc[data[key] > mean + u_scale, key] = drop_val

    _print('Min-Max value (post outlier filtering):',
           np.nanmin(data[key][data[key] != drop_val]),
           np.nanmax(data[key][data[key] != drop_val]))

    return data


def filter_float_df_by_outlier(data, filter_outlier_percent,
                               drop_val=999):

    # For length of code / readability
    fop = proc_fop(filter_outlier_percent)

    if fop[0] is not None:
        data[data < data.quantile(fop[0])] = drop_val
    if fop[1] is not None:
        data[data > data.quantile(fop[1])] = drop_val

    return data


def filter_float_df_by_std(data, n_std,
                           drop_val=999):

    if not isinstance(n_std, tuple):
        n_std = (n_std, n_std)

    mean = data.mean()
    std = data.std()

    if n_std[0] is not None:
        l_scale = n_std[0] * std
        data[data < mean - l_scale] = drop_val

    if n_std[1] is not None:
        u_scale = n_std[1] * std
        data[data > mean + u_scale] = drop_val

    return data


def get_unique_combo_df(data, keys):
    '''Get the unique label combinations from a dataframe (data)
    given multiple column names.

    Parameters
    ----------
    data : pandas DataFrame
        BPt formatted df. Must contain columns with `keys`.

    keys : list
        Column keys within `data`, to merge unique values for.

    Returns
    ----------
    pandas Series
        Series as indexed by subject (BPt) format,
        containing the merged unique values.
    '''

    combo = [data[k].astype(str) + '***' for k in keys]
    combo = reduce(add, combo).dropna()

    label_encoder = LabelEncoder()
    combo[data.index] = label_encoder.fit_transform(combo)

    return combo, label_encoder


def get_unique_combo(to_join):

    as_str = ['***'.join(to_join[i].astype(str)) for i in range(len(to_join))]
    le = LabelEncoder()

    unique_combo = le.fit_transform(as_str)
    return unique_combo, le


def reverse_unique_combo(unique_combo, le):

    reverse = le.inverse_transform(unique_combo)
    seperate = np.array([np.array(reverse[i].split('***')).astype(float)
                        for i in range(len(reverse))])

    return seperate


def reverse_unique_combo_df(unique_combo, le):

    reverse = le.inverse_transform(unique_combo)
    col_split = np.array([r.split('***')[:-1] for r in reverse])
    col_split = col_split.astype(float).astype(int)

    return col_split


def drop_duplicate_cols(data, corr_thresh, _print=print):
    '''Drop duplicates columns within data based on
    if two data columns are >= to a certain correlation threshold.

    Parameters
    ----------
    data : pandas DataFrame
        BPt formatted df.

    corr_thresh : float
        A value between 0 and 1, where if two columns within .data
        are correlated >= to `corr_thresh`, the second column is removed.

        A value of 1 will instead make a quicker direct =='s comparison.

    Returns
    ----------
    pandas DataFrame
        BPt formatted df with duplicates removes

    list
        The list of columns which were dropped from `data`
    '''

    if corr_thresh is not None and corr_thresh is not False:

        dropped = []

        for col1 in data:
            for col2 in data:
                if col1 != col2 and col1 in list(data) and col2 in list(data):

                    A, B = data[col1], data[col2]
                    a, b = ma.masked_invalid(A), ma.masked_invalid(B)

                    # Compute overlap
                    overlap = (~a.mask & ~b.mask)

                    # Two cases, if corr_thresh 1 (faster)
                    if corr_thresh == 1:

                        A_o, B_o = np.array(A[overlap]), np.array(B[overlap])
                        if (A_o == B_o).all():

                            data = data.drop(col2, axis=1)
                            dropped.append(col2)

                    else:
                        corr = np.corrcoef(A[overlap], B[overlap])[0][1]
                        if corr >= corr_thresh:
                            data = data.drop(col2, axis=1)
                            dropped.append(col2)

        _print('Dropped', len(dropped), 'columns as duplicate cols!')

    return data


def get_r(val):

    if isinstance(val, int):
        return val

    return str(np.round(float(val), 3))


def get_original_cat_names(names, encoder, original_key):

    if isinstance(encoder, dict):
        original = [encoder[name] for name in names]

    # K bins encoder case
    elif hasattr(encoder, 'bin_edges_'):
        bin_edges = encoder.bin_edges_[0]
        names = []

        for i in range(len(bin_edges)-1):
            names.append(get_r(bin_edges[i]) + ' - ' + get_r(bin_edges[i+1]))

        # Check and see if NaNs were passed as the last class
        if hasattr(encoder, 'nan_val'):
            names.append('nan')

        return names

    else:

        try:
            float(names[0])
            base = names
        except ValueError:
            base = [int(float(name.replace(original_key + '_', '')))
                    for name in names]

        try:
            original = encoder.inverse_transform(base)
        except IndexError:
            original = encoder.inverse_transform([int(b) for b in base])

    return original


def substrs(x):
    return {x[i:i+j] for i in range(len(x)) for j in range(len(x) - i + 1)}


def find_substr(data):

    s = substrs(data[0])

    for val in data[1:]:
        s.intersection_update(substrs(val))

    try:
        mx = max(s, key=len)

    except ValueError:
        mx = ''

    return mx


def get_top_substrs(keys):

    found = []
    top = find_substr(keys)

    while len(top) > 1:
        found.append(top)

        keys = [k.replace(top, '') for k in keys]
        top = find_substr(keys)

    return found


def proc_datatypes(data_types, col_names):

    # For multi-label case
    if data_types == 'multilabel' or data_types == 'm':
        col_names = [col_names]

    if not isinstance(data_types, list):
        data_types = list([data_types])

    if len(data_types) == 1:
        data_types = [data_types[0] for i in range(len(col_names))]

    if len(data_types) != len(col_names):
        raise RuntimeError('The same number of datatypes were not passed as',
                           'columns!')

    return data_types, col_names


def proc_args(args, data_types):

    if not isinstance(args, list):
        args = list([args])

    # Set to same arg for all if only one passed
    if len(args) == 1:
        args = [args[0] for i in range(len(data_types))]

    if len(args) != len(data_types):
        raise RuntimeError('The length of', args, 'must match length of input',
                           'cols!')

    return args


def process_multilabel_input(keys):

    if not isinstance(keys, list):
        raise RuntimeError(keys, 'must be a list for',
                           'multilabel type')

    if len(keys) < 2:
        raise RuntimeError(keys, 'must be read from multiple',
                           'columns for multilabel type')

    return get_common_name(keys)


def get_common_name(keys):

    top_strs = get_top_substrs(keys)
    if len(top_strs) == 0:
        return keys[0]
    else:
        return top_strs[0]


def filter_data_cols(data, filter_outlier_percent, filter_outlier_std,
                     drop_or_na='drop', seperate_keys=None,
                     subject_id='subject_id', _print=print):

    # Seperate data from data files if applicable
    if seperate_keys is not None:
        sep_data = data[seperate_keys]
        data = data.drop(seperate_keys, axis=1)
    else:
        sep_data = pd.DataFrame()

    if filter_outlier_percent is None and filter_outlier_std is None:
        return data

    if filter_outlier_percent is not None and filter_outlier_std is not None:
        raise RuntimeError('You may only pass one of filter outlier',
                           ' percent or std')

    if drop_or_na == 'na':
        drop_val = np.nan
    else:
        drop_val = get_unused_drop_val(data)

    # Filter based on outlier percent
    if filter_outlier_percent is not None:

        data = filter_float_df_by_outlier(data, filter_outlier_percent,
                                          drop_val=drop_val)

    # Filter by std
    if filter_outlier_std is not None:

        data = filter_float_df_by_std(data, filter_outlier_std,
                                      drop_val=drop_val)

    # Only remove if not NaN
    if drop_val is not np.nan:
        data = drop_from_filter(data, drop_val, _print=_print)

    # Re-merge, if seperated data files
    if seperate_keys is not None:
        data = pd.merge(data, sep_data, on=subject_id)

    return data


def filter_data_file_cols(data, reduce_funcs, filter_outlier_percent,
                          filter_outlier_std, data_file_keys,
                          file_mapping, subject_id='subject_id',
                          n_jobs=1, _print=print):

    if not isinstance(reduce_funcs, list):
        reduce_funcs = [reduce_funcs]

    # Data file proxies returns a list of dataframes
    # where the length of the list is == to the number of reduce funcs
    data_file_proxies = load_data_file_proxies(data, reduce_funcs,
                                               data_file_keys,
                                               file_mapping, n_jobs)

    valid_subjects = set(data.index)
    for proxy in data_file_proxies:

        # drop_or_na is fixed as drop.
        proxy = filter_data_cols(proxy, filter_outlier_percent,
                                 filter_outlier_std, drop_or_na='drop',
                                 seperate_keys=None, subject_id=subject_id,
                                 _print=_print)

        # Only need to save from each calculation to the valid subjects
        valid_subjects = valid_subjects.intersection(proxy.index)

    filtered_data = data.loc[list(valid_subjects)]

    # Remove any unused file mapping entries
    all_keys = set(list(file_mapping))
    remaining_keys = set(np.unique(filtered_data))

    # Calculate the keys to remove, and remove them
    remove_keys = all_keys - remaining_keys
    [file_mapping.pop(key) for key in remove_keys]

    return filtered_data, file_mapping


def drop_from_filter(data, drop_val=999, _print=print):

    to_drop = data[(data == drop_val).any(axis=1)].index

    if len(to_drop) > 0:
        data = data.drop(to_drop)
        _print('Dropped', len(to_drop), 'rows based on filter input',
               'params, e.g. filter outlier percent, drop cat, ect...')
    return data


def proc_file_input(files, file_to_subject, df, subject_id):

    if files is None:
        return df

    if not isinstance(files, dict):
        raise ValueError('files must be passed as a python dict')

    if file_to_subject is None:
        raise RuntimeError('file_to_subject must be specified!')

    if not isinstance(file_to_subject, dict):
        file_to_subject = {key: file_to_subject for key in files}

    for key in files:
        if key not in file_to_subject:
            raise ValueError('If passing file_to_subject as a dict '
                             'then a value must be passed for all '
                             'keys in files. ' + repr(key) + 'was '
                             'not passed in this case.')

    # Compute pd series for each passed
    files_series, all_subjects = dict(), []
    for key in files:

        file_paths = files[key]
        subjects = [file_to_subject[key](fp) for fp in file_paths]
        files_series[key] = pd.Series(file_paths, index=subjects)
        all_subjects += subjects

    # Compute the overlap of subjects
    subjects = sorted(list(set(all_subjects)))

    # Make a dataframe + fill
    files_df = pd.DataFrame(index=subjects)
    files_df.index.name = subject_id
    for key in files_series:
        files_df[key] = files_series[key]

    # If no df, then return
    if df is None:
        return files_df

    # Merge with passed df, and return
    if not isinstance(df, list):
        df = [df]
    df.append(files_df)

    return df


def get_target_type(target_series):

    if target_series.dtype.name == 'category':
        if len(target_series.dtype.categories) == 2:
            return 'binary'
        else:
            return 'categorical'

    return 'regression'



from ..helpers.ML_Helpers import (get_obj_and_params, update_mapping,
                                  proc_mapping, get_reverse_mapping)
import numpy as np
from ..extensions.Loaders import Identity, SurfLabels
from joblib import Parallel, delayed
import warnings
from sklearn.utils.validation import check_memory
from sklearn.base import clone
from .ScopeObjs import ScopeTransformer


def load_and_trans(transformer, load_func, loc):
    '''This function is designed to be able to be wrapped in a cache
    check_memory.'''

    data = load_func(loc)
    trans_data = np.squeeze(transformer.fit_transform(data))
    return trans_data


def get_trans_chunk(transformer, data_files, func):
    '''This function is designed to be used for multi-processing'''

    X_trans_chunk = []
    for data_file in data_files:
        loc = data_file.loc
        load_func = data_file.load_func
        trans_data = func(clone(transformer), load_func, loc)
        X_trans_chunk.append(trans_data)

    return X_trans_chunk


class BPtLoader(ScopeTransformer):

    # Override
    _required_parameters = ["estimator", "inds", "file_mapping"]

    def __init__(self, estimator, inds, file_mapping,
                 n_jobs=1, fix_n_jobs=False,
                 cache_loc=None):

        # Set Super params
        super().__init__(estimator=estimator, inds=inds, cache_loc=cache_loc)

        # Set rest of params
        self.file_mapping = file_mapping

        # Make sure to set fix n jobs before n_jobs
        self.fix_n_jobs = fix_n_jobs
        self.n_jobs = n_jobs

    @property
    def _n_jobs(self):

        if self.fix_n_jobs is False:
            return self.n_jobs

        return self.fix_n_jobs

    def fit(self, X, y=None, mapping=None,
            train_data_index=None, **fit_params):

        # Need the output from a transform to full fit,
        # so when fit is called, call fit_transform instead
        self.fit_transform(X=X, y=y, mapping=mapping,
                           train_data_index=train_data_index,
                           **fit_params)

        return self

    def _fit(self, X, y=None, **fit_params):
        '''Special fit for loader, only concerned with
        fitting first data point for e.g., use of reverse
        transform
        '''

        # Get the first data point
        fit_fm_key = X[0, self.inds_[0]]
        fit_X = self.file_mapping[int(fit_fm_key)].load()

        # Fit the first data point
        self.estimator_.fit(fit_X, y=y, **fit_params)

    def fit_transform(self, X, y=None, mapping=None,
                      train_data_index=None, **fit_params):

        if mapping is None:
            mapping = {}

        # Call parent fit but passing only the first data point
        super().fit(X, y=y, mapping=mapping,
                    train_data_index=train_data_index,
                    **fit_params)

        # The parent fit takes care of, in addition to
        # fitting the loader on one
        # data point, sets base_dtype, processes the mapping,
        # sets rest inds, etc...

        # Now transform X - this sets self.X_trans_inds_
        X_trans = self.transform(X)

        # Need to set out_mapping_ and update mapping
        self.out_mapping_ = {}

        # Add changed X_trans by col
        for c in range(len(self.inds_)):
            ind = self.inds_[c]
            self.out_mapping_[ind] = self.X_trans_inds_[c]

        # Fill the remaining spots sequentially,
        # for each of the rest inds.
        for c in range(len(self.rest_inds_)):
            ind = self.rest_inds_[c]
            self.out_mapping_[ind] = self.n_trans_feats_ + c

        # Update the original mapping, this is the mapping which
        # will be passed to the next piece of the pipeline
        update_mapping(mapping, self.out_mapping_)

        # Now return X_trans
        return X_trans

    def transform(self, X):

        # Init lists + mappings
        X_trans, self.X_trans_inds_ = [], []
        cnt = 0

        # For each column to fit_transform
        for col in self.inds_:

            # Get transformer column
            fm_keys = [key for key in X[:, col]]
            X_trans_cols = self._get_trans_col(fm_keys)

            # Stack + append new features
            X_trans_cols = np.stack(X_trans_cols)
            X_trans.append(X_trans_cols)

            # Add + append inds
            X_trans_cols_inds =\
                [i for i in range(cnt, X_trans_cols.shape[1] + cnt)]
            self.X_trans_inds_.append(X_trans_cols_inds)

            # Update cnt
            cnt = X_trans_cols.shape[1] + cnt

        # Stack final
        X_trans = np.hstack(X_trans)

        # Save number of output features after X_trans
        self.n_trans_feats_ = X_trans.shape[1]

        # Return stacked X_trans with rest inds
        return np.hstack([X_trans, X[:, self.rest_inds_]])

    def get_chunks(self, data_files):

        per_chunk = len(data_files) // self._n_jobs
        chunks = [list(range(i * per_chunk, (i+1) * per_chunk))
                  for i in range(self._n_jobs)]

        last = chunks[-1][-1]
        chunks[-1] += list(range(last+1, len(data_files)))
        return [[data_files[i] for i in c] for c in chunks]

    def _get_trans_col(self, fm_keys):

        # Grab the right data files from the file mapping (casting to int!)
        data_files = [self.file_mapping[int(fm_key)] for fm_key in fm_keys]

        # Clone the base loader
        cloned_estimator = clone(self.estimator)

        # If a caching location is passed, create new load_and_trans_c func
        if self.cache_loc is not None:
            memory = check_memory(self.cache_loc)
            load_and_trans_c = memory.cache(load_and_trans)
        else:
            load_and_trans_c = load_and_trans

        if self._n_jobs == 1:
            X_trans_cols = get_trans_chunk(cloned_estimator,
                                           data_files, load_and_trans_c)
        else:
            chunks = self.get_chunks(data_files)

            X_trans_chunks =\
                Parallel(n_jobs=self._n_jobs)(
                    delayed(get_trans_chunk)(
                        transformer=cloned_estimator,
                        data_files=chunk,
                        func=load_and_trans_c)
                    for chunk in chunks)

            X_trans_cols = []
            for chunk in X_trans_chunks:
                X_trans_cols += chunk

        return X_trans_cols

    def transform_df(self, df, base_name='loader'):

        return super().transform_df(df, base_name=base_name)

    def _proc_new_names(self, feat_names, base_name):

        # Get new names
        new_names = []
        for c in range(len(self.inds_)):

            ind = self.inds_[c]
            base_name = feat_names[ind]

            new_inds = self.X_trans_inds_[c]
            new_names += [base_name + '_' + str(i)
                          for i in range(len(new_inds))]

        # Remove old names - using parent method
        feat_names = self._remove_old_names(feat_names)

        # New names come first, then rest of names
        all_names = new_names + feat_names

        return all_names

    def inverse_transform(self, X, name='base loader'):

        # For each column, compute the inverse transform of what's loaded
        inverse_X = {}
        reverse_mapping = get_reverse_mapping(self.mapping_)

        no_it_warns = set()
        other_warns = set()

        for col_ind in self.inds_:
            reverse_inds = proc_mapping([col_ind], self.out_mapping_)

            # for each subject
            X_trans = []
            for subject_X in X[:, reverse_inds]:

                # If pipeline
                if hasattr(self.estimator_, 'steps'):
                    for step in self.estimator_.steps[::-1]:
                        s_name = step[0]
                        pipe_piece = self.estimator_[s_name]

                        try:
                            subject_X = pipe_piece.inverse_transform(subject_X)
                        except AttributeError:
                            no_it_warns.add(name + '__' + s_name)
                        except:
                            other_warns.add(name + '__' + s_name)

                else:
                    try:
                        subject_X =\
                            self.estimator_.inverse_transform(
                                subject_X)
                    except AttributeError:
                        no_it_warns.add(name)
                    except:
                        other_warns.add(name)

                # Regardless of outcome, add to X_trans
                X_trans.append(subject_X)

            # If X_trans only has len 1, get rid of internal list
            if len(X_trans) == 1:
                X_trans = X_trans[0]

            # Store the list of inverse_transformed X's by subject
            # In a dictionary with the original col_ind as the key
            inverse_X[reverse_mapping[col_ind]] = X_trans

        # Send out warn messages
        if len(no_it_warns) > 0:
            warnings.warn(repr(list(no_it_warns)) + ' skipped '
                          'in calculating inverse FIs due to no '
                          'inverse_transform')
        if len(other_warns) > 0:
            warnings.warn(repr(list(other_warns)) + ' skipped '
                          'in calculating inverse FIs due to '
                          'an error in inverse_transform')

        # Now need to do two things, it is assumed the output from loader
        # cannot be put in a standard X array, but also
        # in the case with multiple loaders, we still need to return
        # An otherwise inversed X, we will just set values to 0 in this version
        reverse_rest_inds = proc_mapping(self.rest_inds_, self.out_mapping_)

        all_inds_len = len(self.inds_) + len(self.rest_inds_)
        Xt = np.zeros((X.shape[0], all_inds_len), dtype=X.dtype)

        Xt[:, self.inds_] = 0
        Xt[:, self.rest_inds_] = X[:, reverse_rest_inds]

        return Xt, inverse_X


LOADERS = {
    'identity': (Identity, ['default']),
    'surface rois': (SurfLabels, ['default']),
}

# If nilearn dependencies
try:
    from nilearn.input_data import NiftiLabelsMasker
    from ..extensions.Loaders import Connectivity
    LOADERS['volume rois'] = (NiftiLabelsMasker, ['default'])
    LOADERS['connectivity'] = (Connectivity, ['default'])

except ImportError:
    pass


def get_loader_and_params(loader_str, extra_params, params,
                          random_state=None, num_feat_keys=None):

    loader, extra_loader_params, loader_params =\
        get_obj_and_params(loader_str, LOADERS, extra_params, params)

    return loader(**extra_loader_params), loader_params



import pandas as pd
import numpy as np

from .Perm_Feat_Importance import Perm_Feat_Importance
from sklearn.inspection import permutation_importance
from ..helpers.ML_Helpers import get_obj_and_params

import seaborn as sns
import math
import matplotlib.pyplot as plt


def get_top_global(df, top_n, get_abs):

    if get_abs:
        imps = np.mean(np.abs(df))
    else:
        imps = np.mean(df)

    imps.sort_values(ascending=False, inplace=True)

    to_get = list(range(top_n))
    if not get_abs:
        to_get += [-i for i in range(1, top_n+1)][::-1]

    top = imps[to_get]
    feats = imps[to_get].index

    top_df = df[feats]

    if get_abs:
        top_df = top_df.abs()

    wide = top_df.melt()
    wide['mean'] = wide['variable'].copy()
    wide['mean'] = wide['mean'].replace(top.to_dict())

    return wide


class Feat_Importances():

    def __init__(self, importance_info, params, n_jobs, scorer):

        self.name = importance_info['name']
        self.scopes = importance_info['scopes']
        self.split = importance_info['split']

        # Unpack params
        self.shap_params = params.shap_params
        self.n_perm = params.n_perm

        self.inverse_global = params.inverse_global
        self.inverse_local = params.inverse_local

        self.valid = True
        self.test = False

        self.n_jobs = n_jobs
        self.scorer = scorer

        self.global_df = None
        self.local_df = None
        self.local_dfs = []

        self.inverse_global_fis = []
        self.inverse_local_fis = []
        self.warning = False

    def get_name(self):

        name = self.name[0].upper() + self.name[1:]
        return name

    def get_data_needed_flags(self, flags):

        # Save flag info
        self.flags = flags

        # Proc by combination of flags and type of feat importance
        if self.name == 'base':

            if flags['tree'] or flags['linear']:
                return False
            else:
                self.valid = False
                return False

        # If shap, return True
        elif self.name == 'shap':

            tfp =\
                self.shap_params.tree_feature_perturbation == 'interventional'
            if flags['tree'] and tfp:
                return False
            return True

    def init_global(self, X, y):

        if not self.valid:
            return

        if 'global' in self.scopes:

            feat_names = list(X)
            self.global_df = pd.DataFrame(columns=feat_names)

    def init_local(self, X, y, test=False, n_splits=None):

        if not self.valid:
            return

        if test:
            self.test = True

        if 'local' in self.scopes:

            # If test or, split as test,
            if self.split == 'test' or self.test:

                self.local_df = X.copy()
                for col in self.local_df.columns:
                    self.local_df[col].values[:] = 0

            # Train or both
            else:

                self.local_df = []
                for i in range(n_splits):

                    df = X.copy()
                    for col in df.columns:
                        df[col].values[:] = 0

                    self.local_df.append(df)

    def get_avg(self, dfs):

        if isinstance(dfs, list):

            df = dfs[0].copy()

            df_arrays = [np.array(df) for df in dfs]
            mean_array = np.nanmean(df_arrays, axis=0)

            # Set to average
            df[list(df)] = mean_array
            return df

        else:
            return dfs

    def add_to_global(self, feat_names, feat_imps):

        feat_imp_dict = {name: imp for name, imp in zip(feat_names, feat_imps)}

        self.global_df = self.global_df.append(feat_imp_dict,
                                               ignore_index=True)
        self.global_df = self.global_df.fillna(0)

    def _add_to_local(self, X_test, vals, fold, class_df):

        val_df = X_test.copy()
        val_df[list(X_test)] = vals

        if self.split == 'test':
            class_df.update(val_df)

        elif self.split == 'train':
            class_df[fold].update(val_df)

            # Replace any rows with all zeros as NaN instead of 0
            ind_df = self.local_df[fold]
            to_nan = ind_df[ind_df.sum(axis=1) == 0].index
            class_df[fold].loc[to_nan] = np.nan

        elif self.split == 'all':
            class_df[fold].update(val_df)

    def add_to_local(self, X_test, vals, fold):
        self._add_to_local(X_test, vals, fold, self.local_df)

    def proc_local(self):

        if not self.valid:
            return

        if 'local' in self.scopes:

            # Get avg then save in self.local_dfs
            df = self.get_avg(self.local_df)
            self.local_dfs.append(df)

            # Reset to None once added
            self.local_df = None

    def proc_importances(self, base_model, X_test, y_test=None,
                         X_train=None, fold=0, random_state=None):
        '''X_test should be a df, and X_train either None or as np array.'''

        if not self.valid:
            return None, None

        if self.name == 'base':
            feat_imps = self.get_base_feat_importances(base_model)
            self.add_to_global(list(X_test), feat_imps)
            return feat_imps, None

        elif self.name == 'perm':
            feat_imps = self.get_perm_feat_importances(base_model,
                                                       np.array(X_test),
                                                       y_test)
            self.add_to_global(list(X_test), feat_imps)
            return feat_imps, None

        elif self.name == 'sklearn perm':
            feat_imps = self.get_perm_feat_importances2(base_model,
                                                        np.array(X_test),
                                                        y_test,
                                                        random_state)
            self.add_to_global(list(X_test), feat_imps)
            return feat_imps, None

        elif self.name == 'shap':
            shap_vals = self.get_shap_feature_importance(base_model, X_test,
                                                         X_train)
            global_shap_vals = self.global_from_local(shap_vals)

            # Add to local
            self.add_to_local(X_test, shap_vals, fold)

            # Add to global
            self.add_to_global(list(X_test), global_shap_vals)
            return global_shap_vals, shap_vals

    def global_from_local(self, vals):
        return self.col_abs_mean(vals)

    def col_abs_mean(self, vals):

        if self.shap_params.avg_abs:
            col_means = np.mean(np.abs(vals), axis=0)

        else:
            col_means = np.mean(vals, axis=0)

        return col_means

    def get_base_feat_importances(self, base_model):

        if self.flags['linear']:
            feat_imps = np.squeeze(base_model.coef_)
        elif self.flags['tree']:
            feat_imps = base_model.feature_importances_

        return feat_imps

    def get_perm_feat_importances(self, base_model, X_test,
                                  y_test):

        perm_import = Perm_Feat_Importance(n_perm=self.n_perm,
                                           n_jobs=self.n_jobs)
        feat_imps = perm_import.compute(base_model, self.scorer,
                                        X_test, y_test)
        return feat_imps

    def get_perm_feat_importances2(self, base_model, X_test, y_test,
                                   random_state):

        results = permutation_importance(base_model, X_test, y_test,
                                         scoring=self.scorer,
                                         n_repeats=self.n_perm,
                                         n_jobs=self.n_jobs,
                                         random_state=random_state)
        return results.importances_mean

    def get_shap_feature_importance(self, base_model, X_test, X_train):

        try:
            import shap
        except ImportError:
            raise ImportError('You must have shap installed to use shap')

        if self.flags['tree'] or self.flags['linear']:

            if self.flags['linear']:

                fp = self.shap_params.linear_feature_perturbation
                n = self.shap_params.linear_nsamples
                explainer = shap.LinearExplainer(base_model, X_train,
                                                 nsamples=n,
                                                 feature_perturbation=fp)

                shap_values = explainer.shap_values(X_test)

            elif self.flags['tree']:

                tmo = self.shap_params.tree_model_output
                tfp = self.shap_params.tree_feature_perturbation
                explainer =\
                    shap.TreeExplainer(
                        base_model, X_train,
                        model_output=tmo,
                        feature_perturbation=tfp)

                ttl = self.shap_params.tree_tree_limit
                shap_values =\
                    explainer.shap_values(X_test,
                                          tree_limit=ttl)

        # Kernel
        else:

            nkmean = self.shap_params.kernel_nkmean

            if nkmean is not None:
                X_train_summary = shap.kmeans(X_train, nkmean)
            else:
                X_train_summary = X_train

            explainer =\
                self.get_kernel_explainer(base_model, X_train_summary,
                                          self.shap_params.kernel_link)

            klr = self.shap_params.kernel_l1_reg
            kns = self.shap_params.kernel_nsamples

            shap_values =\
                explainer.shap_values(np.array(X_test),
                                      l1_reg=klr,
                                      n_samples=kns)

        return self.proc_shap_vals(shap_values)

    def proc_shap_vals(self, shap_values):
        return shap_values

    def get_kernel_explainer(self, model, X_train_summary, link):

        try:
            import shap
        except ImportError:
            raise ImportError('You must have shap installed to use shap')

        if link == 'default':
            link = 'logic'

        explainer = shap.KernelExplainer(model.predict_proba, X_train_summary,
                                         link=link)

        return explainer

    def set_final_local(self):

        if 'local' in self.scopes:

            df = self.get_avg(self.local_dfs)
            self.local_df = df

        self.local_dfs = None  # Save memory

    def get_global_label(self):

        if self.name == 'base':
            if self.flags['linear']:
                return 'Mean Beta Weight'
            elif self.flags['tree']:
                return 'Mean Tree Split Based Importance'

        elif self.name == 'shap':
            return 'Mean Shap Value'

    def _get_warning_str(self, as_html=False):

        lb = '\n'
        if as_html:
            lb = '<br>'

        content = 'Warning! Base pipeline contains either loaders '
        content += 'or transformers. Any self.global_df or self.local_df '
        content += 'might therefore be incorrect, or atleast hard to '
        content += 'make sense of. In this case you should likely use the '
        content += 'the inverse feat importances!' + lb

        return content

    def _repr_html_(self):

        content = 'Feature_Importances Object<br>'

        if not self.valid:
            return content + 'Invalid type for passed classifier!<br>'

        content += 'name: ' + self.name + '<br>'

        if self.name == 'base':

            content += 'importance type: '
            if self.flags['linear']:
                content += 'beta weight <br>'
            elif self.flags['tree']:
                content += 'tree based <br>'

        if self.warning:

            content += self._get_warning_str(as_html=True)

            if len(self.inverse_global_fis) > 0:
                content += 'contains inverse global feat importances: '
                content += 'self.inverse_global_fis <br>'
            if len(self.inverse_local_fis) > 0:
                content += 'contains inverse local feat importances: '
                content += 'self.inverse_local_fis <br>'

        else:
            if 'global' in self.scopes:
                content += 'contains global feat importances: '
                content += 'self.global_df <br>'
            if 'local' in self.scopes:
                content += 'contains local feat importances: '
                content += 'self.local_df <br>'

        content += 'importance computed on: ' + self.split + '<br>'
        return content

    def set_target(self, target):
        self.target = target

    def set_run_name(self, run_name):
        self.run_name = run_name

    def plot_global(self, top_n=10, show_abs=False, multiclass=False,
                    ci=95, palette='default',
                    figsize=(10, 10), title='default', titles='default',
                    xlabel='default', n_cols=1, ax=None, show=True):
        '''Plots any global feature importance, e.g. base or shap, values per
        feature not per prediction.

        Parameters
        ----------
        top_n : int, optional
            The number of top features to display. In the case where
            show_abs is set to True, or the feature importance being plotted
            is only positive, then top_n features will be shown. On the other
            hand, when show_abs is set to False and the feature importances being
            plotted contain negative numbers, then the top_n highest and top_n
            lowest features will be shown.

            (default = 10)

        show_abs : bool, optional
            In the case where the underlying feature importances contain
            negative numbers, you can either plot the top_n by absolute value,
            with show_abs set to True, or plot the top_n highest and lowest
            with show_abs set to False.

            (default = False)

        multiclass : bool, optional
            If multiclass is set to True, and the underlying
            feature importances
            were derived from a categorical problem type, then
            a seperate feature importance plot will be made for each class.
            Alternatively, if multiclass is set to False,
            then feature importances
            will be averaged over all classes.

            (default = False)

        ci : float, 'sd' or None, optional
            Size of confidence intervals to draw around estimated values.
            If 'sd', skip bootstrapping and draw the standard deviation
            of the feat importances. If None, no
            bootstrapping will be performed,
            and error bars will not be drawn.

            (default = 95)

        palette : Seaborn palette name, optional
            Color scheme to use. Search seaborn palettes for more information.
            Default for absolute is 'Reds', and default for both pos
            and neg is 'coolwarm'.

            (default = 'default')

        title : str, optional
            The title used during plotting, and also used
            to save a version of the figure
            (with spaces in title replaced by _, and as a png).

            When multiclass is True, this is the full figure title.

            (default = 'default')

        titles: list, optional
            This parameter is only used when multiclass is True.
            titles should be a list with the name for each classes plot.
            If left as default, it will just be named the original loaded name
            for that class.

            (default = 'default')

        xlabel : str, optional
            The xlabel, descriping the measure of feature importance.
            If left as 'default' it will change depend on
            what feature importance
            is being plotted.

            (default = 'default')

        n_cols: int, optional
            If multiclass, then the number of class plots to
            plot on each row.

            (default = 1)

        ax: matplotlib axis, or axes, optional
            A custom ax to plot to for an individual plot, or if using
            multiclass, then a list of axes can be passed here.

            (default = None)

        show : bool, optional
            If True, then plt.show(), the matplotlib command will be called,
            and the figure displayed. On the other hand, if set to False,
            then the user can customize the plot as they desire.
            You can think of plt.show() as clearing all of the loaded settings,
            so in order to make changes, you can't call
            this until you are done.

            (default = True)
        '''

        # Check
        if self.global_df is None:
            raise AttributeError('No global feature importances found!')

        # If multiclass
        if isinstance(self.global_df, list):
            if multiclass:
                dfs = self.global_df
                return self._plot_multiclass_global(dfs, top_n=top_n,
                                                    show_abs=show_abs,
                                                    ci=ci, palette=palette,
                                                    figsize=figsize,
                                                    title=title,
                                                    titles=titles,
                                                    xlabel=xlabel,
                                                    n_cols=n_cols, ax=ax,
                                                    show=show)
            else:
                df = pd.concat(self.global_df)
        else:
            df = self.global_df

        # Non-multi-class case
        ax = self._plot_global(df, top_n=top_n, show_abs=show_abs,
                               ci=ci, palette=palette, title=title,
                               xlabel=xlabel, ax=ax, show=show)
        return ax

    def _plot_multiclass_global(self, dfs, top_n=10, show_abs=False,
                                ci=95, palette='default',
                                figsize=(10, 10), title='default',
                                titles='default',
                                xlabel='default', n_cols=2,
                                ax=None, show=True):

        # Grab the right classes
        # target = feat_importances.target
        # target_key = self._get_targets_key(target)
        # classes = self.targets_encoders[target_key].classes_
        classes = [str(c) for c in self.n_classes]

        # Set default titles
        if titles == 'default':
            titles = classes

        if title == 'default':
            name = self.get_name()
            title = name + ' Top Features by Class'

        # If an ax is passed, then use that instead of making subplots
        n_rows = math.ceil(len(classes) / n_cols)
        if ax is not None:
            axes = ax
            fig = None
        else:
            fig, axes = plt.subplots(n_rows, n_cols, figsize=figsize)

        # In the case where only 1 row or 1 col.
        single = False
        if n_rows == 1 or n_cols == 1:
            single = True

        for i in range(len(dfs)):

            row = i // n_cols
            col = i % n_cols
            rc = (row, col)

            if single:
                rc = i

            # Create the subplot
            axes[rc] = self._plot_global(
                dfs[i], top_n=top_n,
                show_abs=show_abs, ci=ci, palette=palette,
                title=titles[i], xlabel='default', ax=axes[rc])

        # Set any remaining subplots to nothing
        for i in range(len(dfs), n_rows * n_cols):

            row = i // n_cols
            col = i % n_cols
            rc = (row, col)

            if single:
                rc = i

            axes[rc].set_axis_off()

        # Set the title and layout if no custom ax passed
        if ax is None:
            fig.tight_layout()
            fig.subplots_adjust(top=.9, hspace=.25)
            fig.suptitle(title, fontsize="x-large")

            if show:
                plt.show()

        else:
            return axes

    def _plot_global(self, df, top_n=10,
                     show_abs=False, ci=95,
                     palette='default', title='default',
                     xlabel='default', ax=None, show=True):

        # Distinguish between feat importances that have been set to abs
        min_val = np.min(np.min(df))
        add_abs_sign = True

        if min_val >= 0:
            show_abs = True
        else:
            add_abs_sign = False

        # Get df, with only top_n, in wide form for plotting
        top_df = get_top_global(df, top_n, show_abs)

        # Abs / all pos values
        if show_abs:
            if palette == 'default':
                palette = 'Reds'

            if title == 'default':
                name = self.get_name()
                title = name + ' Top ' + str(top_n) + ' Features'

            if xlabel == 'default':
                xlabel = self.get_global_label()

            sns.barplot(x='value', y='variable', hue='mean',
                        data=top_df, orient='h', ci=ci,
                        palette=palette, errwidth=2, ax=ax, dodge=False)

        # Pos + neg values
        else:
            if palette == 'default':
                palette = 'coolwarm'

            if title == 'default':
                name = self.get_name()
                title = name + ' Top ' + str(top_n) + ' +/- Features'

            if xlabel == 'default':
                xlabel = self.get_global_label()

                if add_abs_sign:
                    xlabel = '|' + xlabel + '|'

            sns.barplot(x='value', y='variable', hue='mean',
                        data=top_df, orient='h', ci=ci,
                        palette=palette, errwidth=2, ax=ax,
                        dodge=False)

            # Add an extra line down the center
            if ax is None:
                current_ax = plt.gca()
                current_ax.axvline(0, color='k', lw=.2)
                sns.despine(ax=current_ax, top=True, left=True, bottom=True,
                            right=True)
            else:
                ax.axvline(0, color='k', lw=.2)
                sns.despine(ax=ax, top=True, left=True,
                            bottom=True, right=True)

        # Process rest fo params based on if for axis or global
        if ax is None:
            plt.title(title)
            plt.xlabel(xlabel)
            plt.ylabel('')
            plt.legend().remove()

            if show:
                plt.show()

        else:
            ax.set_title(title)
            ax.set_xlabel(xlabel)
            ax.set_ylabel('')
            ax.get_legend().remove()
            return ax

    def plot_local(self, top_n=10, title='default', titles='default',
                   xlabel='default', one_class=None, show=True):
        '''Plots any local feature importance, e.g. shap, values per
        per prediction.

        Parameters
        ----------
        top_n : int, optional
            The number of top features to display. In the case where
            show_abs is set to True, or the feature importance being plotted
            is only positive, then top_n features will be shown. On the other
            hand, when show_abs is set to False and the
            feature importances being
            plotted contain negative numbers, then the top_n highest and top_n
            lowest features will be shown.

            (default = 10)

        title : str, optional
            The title used during plotting, and also used
            to save a version of the figure
            (with spaces in title replaced by _, and as a png).

            With a multiclass / categorical problem type, this
            is only used if one_class is set. Otherwise, titles are used.

            (default = 'default')

        titles: list, optional
            This parameter is only used with a multiclass problem type.
            titles should be a list with the name for each class to plot.
            If left as default, it will use originally loaded class names.
            for that class.

            (default = 'default')

        xlabel : str, optional
            The xlabel, descriping the measure of feature importance.
            If left as 'default' it will change depend on what
            feature importance
            is being plotted.

            (default = 'default')

        one_class : int or None, optional
            If an underlying multiclass or categorical type, optionally
            provide an int here, corresponding to the single class to
            plot. If left as None, with make plots for all classes.

            (default = None)

        show : bool, optional
            If True, then plt.show(), the matplotlib command will be called,
            and the figure displayed. On the other hand, if set to False,
            then the user can customize the plot as they desire.
            You can think of plt.show() as clearing all of the loaded settings,
            so in order to make changes, you can't call this until
            you are done.

            (default = True)
        '''

        if self.local_df is None:
            raise AttributeError('No local df found!')

        df = self.local_df

        # If multiclass
        if isinstance(df, list):

            # target_key = self._get_targets_key(feat_importances.target)
            # classes = self.targets_encoders[target_key].classes_
            classes = [str(c) for c in self.n_classes]

            if titles == 'default':
                base = ' Shap ' + str(top_n) + ' Features'
                titles = [cl + base for cl in classes]

            if one_class is None:

                for i in range(len(df)):
                    self._plot_shap_summary(df[i], top_n, titles[i],
                                            xlabel, show)

                return

            # If just one class passed, treat as just one class
            else:
                df = df[one_class]

                if title == 'default':
                    title = titles[one_class]

        self._plot_shap_summary(df, top_n, title, xlabel, show)

    def _plot_shap_summary(self, shap_df, top_n, title, xlabel, show):

        try:
            import shap
        except ImportError:
            raise ImportError('You must have shap installed to use shap')

        # Doesn't work right now - will break

        shap_cols, shap_inds = list(shap_df), shap_df.index
        data = self.all_data.loc[shap_inds][shap_cols]

        shap.summary_plot(np.array(shap_df), data,
                          max_display=top_n, show=False)

        if title == 'default':
            title = 'Shap Summary Top ' + str(top_n) + ' Features'

        plt.title(title)

        if xlabel != 'default':
            plt.xlabel(xlabel)

        if show:
            plt.show()


class Regression_Feat_Importances(Feat_Importances):

    def get_kernel_explainer(self, model, X_train_summary, link):

        try:
            import shap
        except ImportError:
            raise ImportError('You must have shap installed to use shap')

        if link == 'default':
            link = 'identity'

        explainer = shap.KernelExplainer(model.predict, X_train_summary,
                                         link=link)

        return explainer


class Cat_Feat_Importances(Feat_Importances):

    def init_global(self, X, y):

        if not self.valid:
            return

        if 'global' in self.scopes:

            self.n_classes = len(np.unique(y))
            self.global_df = []
            feat_names = list(X)

            for j in range(self.n_classes):
                self.global_df.append(pd.DataFrame(columns=feat_names))

    def init_local(self, X, y=None, test=False, n_splits=None):

        if not self.valid:
            return

        if test:
            self.test = True

        if 'local' in self.scopes:

            self.local_df = []
            self.n_classes = len(np.unique(y))

            # If test or, split as test,
            if self.split == 'test' or self.test:

                for j in range(self.n_classes):

                    df = X.copy()

                    for col in df.columns:
                        df[col].values[:] = 0

                    self.local_df.append(df)

            # Train or both
            else:

                for j in range(self.n_classes):
                    dfs = []

                    for i in range(n_splits):

                        df = X.copy()
                        for col in df.columns:
                            df[col].values[:] = 0

                        dfs.append(df)

                    self.local_df.append(dfs)

    def add_to_global(self, feat_names, feat_imps):

        # In cases like base feat importance for trees
        # just set each class to the global values
        if len(feat_names) == len(feat_imps):
            feat_imps = [feat_imps for j in range(self.n_classes)]

        for j in range(self.n_classes):

            feat_imp_dict = {name: imp for name, imp in
                             zip(feat_names, feat_imps[j])}

            self.global_df[j] = self.global_df[j].append(feat_imp_dict,
                                                         ignore_index=True)
            self.global_df[j] = self.global_df[j].fillna(0)

    def add_to_local(self, X_test, vals, fold):

        for j in range(self.n_classes):
            self._add_to_local(X_test, vals[j], fold, self.local_df[j])

    def global_from_local(self, vals):

        global_vals = []

        for j in range(self.n_classes):
            global_vals.append(self.col_abs_mean(vals[j]))

        return global_vals

    def proc_local(self):

        if not self.valid:
            return

        if 'local' in self.scopes:

            dfs = []
            for j in range(self.n_classes):
                dfs.append(self.get_avg(self.local_df[j]))

            self.local_dfs.append(dfs)

            # Reset to None once added
            self.local_df = None

    def set_final_local(self):

        if not self.valid:
            return

        if 'local' in self.scopes:

            if len(self.local_dfs) > 0:

                dfs = []

                for j in range(self.n_classes):
                    class_dfs = [d[j] for d in self.local_dfs]
                    dfs.append(self.get_avg(class_dfs))

                self.local_df = dfs

            else:
                self.local_df = self.local_dfs[0]

        self.local_dfs = None  # Save memory


IMPORTANCES = {
    'base': ({'name': 'base',
              'scopes': ('global'),
              'split': 'train'}, ['default']),

    'shap': ({'name': 'shap',
              'scopes': ('local', 'global'),
              'split': 'test'}, ['base shap']),

    'shap train': ({'name': 'shap',
                    'scopes': ('local', 'global'),
                    'split': 'train'}, ['base shap']),

    'shap all': ({'name': 'shap',
                  'scopes': ('local', 'global'),
                  'split': 'all'}, ['base shap']),

    'perm': ({'name': 'perm',
              'scopes': ('global'),
              'split': 'test'}, ['base perm']),

    'perm train': ({'name': 'perm',
                    'scopes': ('global'),
                    'split': 'train'}, ['base perm']),

    'perm all': ({'name': 'perm',
                  'scopes': ('global'),
                  'split': 'all'}, ['base perm']),

    'sklearn perm': ({'name': 'sklearn perm',
                      'scopes': ('global'),
                      'split': 'test'}, ['base perm']),

    'sklearn perm train': ({'name': 'sklearn perm',
                            'scopes': ('global'),
                            'split': 'train'}, ['base perm']),

    'sklearn perm all': ({'name': 'sklearn perm',
                          'scopes': ('global'),
                          'split': 'all'}, ['base perm']),
}


def get_feat_importances_and_params(params, problem_type, n_jobs, scorer):

    # Search type should always be None
    imp_info, _, _ =\
        get_obj_and_params(params.obj, IMPORTANCES, {}, params.params)

    problem_types = {'binary': Feat_Importances,
                     'regression': Regression_Feat_Importances,
                     'categorical': Cat_Feat_Importances}

    # Get right base object by problem type
    FI = problem_types[problem_type]

    # Return the instance of the feat_importance obj
    return FI(imp_info, params, n_jobs, scorer)
